import autolens as al
import autolens.plot as aplt
from test_autolens.simulate.imaging import instrument_util

# In this tutorial, we'll introduce a new pixelization, called an adaptive-pixelization. This pixelization doesn't use
# uniform grid of rectangular pixels, but instead uses ir'Voronoi' pixels. So, why would we want to do that?
# Lets take another look at the rectangular grid, and think about its weakness.

# Lets quickly remind ourselves of the image, and the 3.0" circular mask we'll use to mask it.
imaging = instrument_util.load_test_imaging(
    data_name="lens_light_dev_vaucouleurs", instrument="vro"
)

mask = al.Mask.circular(
    shape_2d=imaging.shape_2d,
    pixel_scales=imaging.pixel_scales,
    radius=3.0,
    centre=(3.0, 3.0),
)

# The lines of code below do everything we're used to, that is, setup an image and its grid, mask it, trace it
# via a tracer, setup the rectangular mapper, etc.
lens_galaxy = al.Galaxy(
    redshift=0.5,
    light=al.lp.EllipticalDevVaucouleurs(
        centre=(0.0, 0.0), axis_ratio=0.9, phi=45.0, intensity=0.1, effective_radius=1.0
    ),
)

masked_imaging = al.MaskedImaging(imaging=imaging, mask=mask)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy])
fit = al.FitImaging(masked_imaging=masked_imaging, tracer=tracer)
aplt.FitImaging.subplot_fit_imaging(fit=fit)



import autofit as af
import autolens as al


class TestSlam:
    def test__lens_light_tag_for_source_pipeline(self):

        hyper = al.slam.SLaMHyper()
        source = al.slam.SLaMSource()
        light = al.slam.SLaMLight()
        mass = al.slam.SLaMMass()

        slam = al.slam.SLaM(hyper=hyper, source=source, light=light, mass=mass)

        assert slam.lens_light_tag_for_source_pipeline == ""

        slam.set_light_type(light_type="sersic")

        assert slam.lens_light_tag_for_source_pipeline == "__light_sersic"


class TestSLaMHyper:
    def test__hyper_fixed_after_source(self):
        hyper = al.slam.SLaMHyper(hyper_fixed_after_source=False)
        assert hyper.hyper_fixed_after_source_tag == ""

        hyper = al.slam.SLaMHyper(hyper_fixed_after_source=True)
        assert hyper.hyper_fixed_after_source_tag == "_fixed"

    def test__hyper_tag(self):

        hyper = al.slam.SLaMHyper(
            hyper_galaxies=True,
            hyper_image_sky=True,
            hyper_background_noise=True,
            hyper_fixed_after_source=True,
        )

        assert hyper.hyper_tag == "__hyper_galaxies_bg_sky_bg_noise_fixed"

        hyper = al.slam.SLaMHyper(hyper_galaxies=True, hyper_background_noise=True)

        assert hyper.hyper_tag == "__hyper_galaxies_bg_noise"

        hyper = al.slam.SLaMHyper(
            hyper_fixed_after_source=True,
            hyper_galaxies=True,
            hyper_background_noise=True,
        )

        assert hyper.hyper_tag == "__hyper_galaxies_bg_noise_fixed"


class TestSLaMSource:
    def test__lens_light_bulge_only_tag(self):
        source = al.slam.SLaMSource(lens_light_bulge_only=False)
        assert source.lens_light_bulge_only_tag == ""
        source = al.slam.SLaMSource(lens_light_bulge_only=True)
        assert source.lens_light_bulge_only_tag == "__bulge_only"

    def test__tag(self):

        source = al.slam.SLaMSource(
            pixelization=al.pix.Rectangular,
            regularization=al.reg.Constant,
            lens_light_centre=(1.0, 2.0),
            lens_mass_centre=(3.0, 4.0),
            align_light_mass_centre=False,
            no_shear=True,
        )

        source.type_tag = source.inversion_tag

        assert (
            source.tag
            == "source____pix_rect__reg_const__no_shear__lens_light_centre_(1.00,2.00)__lens_mass_centre_(3.00,4.00)"
        )

        source = al.slam.SLaMSource(
            pixelization=al.pix.Rectangular,
            regularization=al.reg.Constant,
            align_light_mass_centre=True,
            number_of_gaussians=1,
            lens_light_bulge_only=True,
        )

        source.type_tag = "test"

        assert (
            source.tag
            == "source__test__gaussians_x1__with_shear__align_light_mass_centre__bulge_only"
        )

    def test__shear(self):

        source = al.slam.SLaMSource(no_shear=False)
        assert source.shear is al.mp.ExternalShear
        source = al.slam.SLaMSource(no_shear=True)
        assert source.shear == None

    def test__align_centre_of_mass_to_light(self):

        mass = af.PriorModel(al.mp.SphericalIsothermal)

        source = al.slam.SLaMSource(align_light_mass_centre=False)

        mass = source.align_centre_of_mass_to_light(mass=mass, light_centre=(1.0, 2.0))

        assert mass.centre.centre_0.mean == 1.0
        assert mass.centre.centre_0.sigma == 0.1
        assert mass.centre.centre_0.mean == 1.0
        assert mass.centre.centre_0.sigma == 0.1

        source = al.slam.SLaMSource(align_light_mass_centre=True)

        mass = source.align_centre_of_mass_to_light(mass=mass, light_centre=(1.0, 2.0))

        assert mass.centre == (1.0, 2.0)

    def test__align_centre_to_lens_light_centre(self):

        light = af.PriorModel(al.mp.SphericalIsothermal)

        source = al.slam.SLaMSource(lens_light_centre=(1.0, 2.0))

        light = source.align_centre_to_lens_light_centre(light=light)

        assert light.centre == (1.0, 2.0)

    def test__align_centre_to_lens_mass_centre(self):

        mass = af.PriorModel(al.mp.SphericalIsothermal)

        source = al.slam.SLaMSource(lens_mass_centre=(1.0, 2.0))

        mass = source.align_centre_to_lens_mass_centre(mass=mass)

        assert mass.centre == (1.0, 2.0)

    def test__remove_disk_from_lens_galaxy(self):

        lens = al.GalaxyModel(
            redshift=0.5, bulge=al.lp.EllipticalSersic, disk=al.lp.EllipticalExponential
        )

        source = al.slam.SLaMSource(lens_light_bulge_only=False)

        lens = source.remove_disk_from_lens_galaxy(lens=lens)

        assert type(lens.disk) is af.PriorModel

        source = al.slam.SLaMSource(lens_light_bulge_only=True)

        lens = source.remove_disk_from_lens_galaxy(lens=lens)

        assert lens.disk is None

    def test__is_inversion(self):

        source = al.slam.SLaMSource()

        source.type_tag = "sersic"
        assert source.is_inversion == False

        source.type_tag = "anything_else"
        assert source.is_inversion == True

    def test__unfix_lens_mass_centre(self):

        mass = af.PriorModel(al.mp.SphericalIsothermal)
        mass.centre = (1.0, 2.0)

        source = al.slam.SLaMSource()

        mass = af.PriorModel(al.mp.SphericalIsothermal)
        source = al.slam.SLaMSource(lens_mass_centre=(5.0, 6.0))

        mass = source.unfix_lens_mass_centre(mass=mass)

        assert mass.centre.centre_0.mean == 5.0
        assert mass.centre.centre_0.sigma == 0.05
        assert mass.centre.centre_1.mean == 6.0
        assert mass.centre.centre_1.sigma == 0.05


class TestSLaMLight:
    def test__tag(self):

        light = al.slam.SLaMLight(align_bulge_disk_elliptical_comps=True)
        light.type_tag = ""

        assert light.tag == "light____align_bulge_disk_ell"

        light = al.slam.SLaMLight(align_bulge_disk_centre=True, disk_as_sersic=True)

        light.type_tag = "lol"

        assert light.tag == "light__lol__align_bulge_disk_centre__disk_sersic"

        light = al.slam.SLaMLight(
            align_bulge_disk_centre=True,
            align_bulge_disk_elliptical_comps=True,
            disk_as_sersic=True,
            number_of_gaussians=2,
        )
        light.type_tag = "test"

        assert light.tag == "light__test__gaussians_x2"


class TestSLaMMass:
    def test__fix_lens_light_tag(self):
        mass = al.slam.SLaMMass(fix_lens_light=False)
        assert mass.fix_lens_light_tag == ""
        mass = al.slam.SLaMMass(fix_lens_light=True)
        assert mass.fix_lens_light_tag == "__fix_lens_light"

    def test__tag(self):

        mass = al.slam.SLaMMass(
            no_shear=True, align_light_dark_centre=True, fix_lens_light=True
        )
        mass.type_tag = ""

        assert mass.tag == "mass____no_shear__align_light_dark_centre__fix_lens_light"

        mass = al.slam.SLaMMass(align_bulge_dark_centre=True)

        mass.type_tag = "test"

        assert mass.tag == "mass__test__with_shear__align_bulge_dark_centre"

        mass = al.slam.SLaMMass(include_smbh=True, smbh_centre_fixed=False)

        mass.type_tag = "test"

        assert mass.tag == "mass__test__with_shear__smbh_centre_free"

    def test__shear_from_previous_pipeline(self):

        mass = al.slam.SLaMMass(no_shear=True)

        assert mass.shear_from_previous_pipeline == None

        mass = al.slam.SLaMMass(no_shear=False)

        assert isinstance(mass.shear_from_previous_pipeline, af.AbstractPromise)



import autolens as al

from test_autolens.simulators.imaging import instrument_util


def simulate__lens_sie__source_smooth(instrument):

    data_name = "lens_sie__source_smooth"

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mp.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.6, elliptical_comps=(0.17647, 0.0)
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.lp.EllipticalSersic(
            centre=(0.0, 0.0),
            elliptical_comps=(-0.055555, 0.096225),
            intensity=0.4,
            effective_radius=0.5,
            sersic_index=1.0,
        ),
    )

    instrument_util.simulate_imaging_from_instrument(
        instrument=instrument,
        data_name=data_name,
        galaxies=[lens_galaxy, source_galaxy],
    )


def simulate__lens_sie__source_cuspy(instrument):

    data_name = "lens_sie__source_cuspy"

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mp.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.6, elliptical_comps=(0.17647, 0.0)
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.lp.EllipticalSersic(
            centre=(0.0, 0.0),
            elliptical_comps=(-0.055555, 0.096225),
            intensity=0.1,
            effective_radius=0.5,
            sersic_index=3.0,
        ),
    )

    instrument_util.simulate_imaging_from_instrument(
        instrument=instrument,
        data_name=data_name,
        galaxies=[lens_galaxy, source_galaxy],
    )


def simulate__lens_sis__source_smooth(instrument):

    data_name = "lens_sis__source_smooth"

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.6),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.lp.EllipticalSersic(
            centre=(0.0, 0.0),
            elliptical_comps=(-0.055555, 0.096225),
            intensity=0.4,
            effective_radius=0.5,
            sersic_index=1.0,
        ),
    )

    instrument_util.simulate_imaging_from_instrument(
        instrument=instrument,
        data_name=data_name,
        galaxies=[lens_galaxy, source_galaxy],
    )


def simulate__lens_sie__source_smooth__offset_centre(instrument):

    data_name = "lens_sie__source_smooth__offset_centre"

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mp.SphericalIsothermal(centre=(2.0, 2.0), einstein_radius=1.2),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.lp.EllipticalSersic(
            centre=(2.0, 2.0),
            elliptical_comps=(-0.055555, 0.096225),
            intensity=0.4,
            effective_radius=0.5,
            sersic_index=1.0,
        ),
    )

    instrument_util.simulate_imaging_from_instrument(
        instrument=instrument,
        data_name=data_name,
        galaxies=[lens_galaxy, source_galaxy],
    )


def simulate__lens_light__source_smooth(instrument):

    data_name = "lens_light__source_smooth"

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        light=al.lp.EllipticalSersic(
            centre=(0.0, 0.0),
            elliptical_comps=(0.1, 0.0),
            intensity=0.5,
            effective_radius=0.8,
            sersic_index=4.0,
        ),
        mass=al.mp.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.6, elliptical_comps=(0.17647, 0.0)
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.lp.EllipticalSersic(
            centre=(0.0, 0.0),
            elliptical_comps=(-0.055555, 0.096225),
            intensity=0.4,
            effective_radius=0.5,
            sersic_index=1.0,
        ),
    )

    instrument_util.simulate_imaging_from_instrument(
        instrument=instrument,
        data_name=data_name,
        galaxies=[lens_galaxy, source_galaxy],
    )


def simulate__lens_light__source_cuspy(instrument):

    data_name = "lens_light__source_cuspy"

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        light=al.lp.EllipticalSersic(
            centre=(0.0, 0.0),
            elliptical_comps=(0.1, 0.0),
            intensity=0.5,
            effective_radius=0.8,
            sersic_index=4.0,
        ),
        mass=al.mp.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.6, elliptical_comps=(0.17647, 0.0)
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.lp.EllipticalSersic(
            centre=(0.0, 0.0),
            elliptical_comps=(-0.055555, 0.096225),
            intensity=0.1,
            effective_radius=0.5,
            sersic_index=3.0,
        ),
    )

    instrument_util.simulate_imaging_from_instrument(
        instrument=instrument,
        data_name=data_name,
        galaxies=[lens_galaxy, source_galaxy],
    )



import autolens as al
import autofit as af


def test__lens_light_centre_tag():

    setup = al.SetupPipeline(lens_light_centre=None)
    assert setup.lens_light_centre_tag == ""
    setup = al.SetupPipeline(lens_light_centre=(2.0, 2.0))
    assert setup.lens_light_centre_tag == "__lens_light_centre_(2.00,2.00)"
    setup = al.SetupPipeline(lens_light_centre=(3.0, 4.0))
    assert setup.lens_light_centre_tag == "__lens_light_centre_(3.00,4.00)"
    setup = al.SetupPipeline(lens_light_centre=(3.027, 4.033))
    assert setup.lens_light_centre_tag == "__lens_light_centre_(3.03,4.03)"


def test__lens_mass_centre_tag():

    setup = al.SetupPipeline(lens_mass_centre=None)
    assert setup.lens_mass_centre_tag == ""
    setup = al.SetupPipeline(lens_mass_centre=(2.0, 2.0))
    assert setup.lens_mass_centre_tag == "__lens_mass_centre_(2.00,2.00)"
    setup = al.SetupPipeline(lens_mass_centre=(3.0, 4.0))
    assert setup.lens_mass_centre_tag == "__lens_mass_centre_(3.00,4.00)"
    setup = al.SetupPipeline(lens_mass_centre=(3.027, 4.033))
    assert setup.lens_mass_centre_tag == "__lens_mass_centre_(3.03,4.03)"


def test__align_light_mass_centre_tag__is_empty_sting_if_both_lens_light_and_mass_centres_input():
    setup = al.SetupPipeline(align_light_mass_centre=False)
    assert setup.align_light_mass_centre_tag == ""
    setup = al.SetupPipeline(align_light_mass_centre=True)
    assert setup.align_light_mass_centre_tag == "__align_light_mass_centre"
    setup = al.SetupPipeline(
        lens_light_centre=(0.0, 0.0),
        lens_mass_centre=(1.0, 1.0),
        align_light_mass_centre=True,
    )
    assert setup.align_light_mass_centre_tag == ""


def test__no_shear_tag():
    setup = al.SetupPipeline(no_shear=False)
    assert setup.no_shear_tag == "__with_shear"

    setup = al.SetupPipeline(no_shear=True)
    assert setup.no_shear_tag == "__no_shear"


def test__constant_mass_to_light_ratio_tag():

    setup = al.SetupPipeline(constant_mass_to_light_ratio=True)
    assert setup.constant_mass_to_light_ratio_tag == "__mlr_const"
    setup = al.SetupPipeline(constant_mass_to_light_ratio=False)
    assert setup.constant_mass_to_light_ratio_tag == "__mlr_free"


def test__align_light_dark_tag():

    setup = al.SetupPipeline(align_light_dark_centre=False)
    assert setup.align_light_dark_centre_tag == ""
    setup = al.SetupPipeline(align_light_dark_centre=True)
    assert setup.align_light_dark_centre_tag == "__align_light_dark_centre"


def test__align_bulge_dark_tag():
    setup = al.SetupPipeline(align_bulge_dark_centre=False)
    assert setup.align_bulge_dark_centre_tag == ""
    setup = al.SetupPipeline(align_bulge_dark_centre=True)
    assert setup.align_bulge_dark_centre_tag == "__align_bulge_dark_centre"


def test__smbh_tag():

    setup = al.SetupPipeline(include_smbh=False)
    assert setup.include_smbh_tag == ""

    setup = al.SetupPipeline(include_smbh=True, smbh_centre_fixed=True)
    assert setup.include_smbh_tag == "__smbh_centre_fixed"

    setup = al.SetupPipeline(include_smbh=True, smbh_centre_fixed=False)
    assert setup.include_smbh_tag == "__smbh_centre_free"


def test__subhalo_centre_tag():

    setup = al.SetupPipeline(subhalo_instance=None)
    assert setup.subhalo_centre_tag == ""
    setup = al.SetupPipeline(subhalo_instance=al.mp.SphericalNFW(centre=(2.0, 2.0)))
    assert setup.subhalo_centre_tag == "__sub_centre_(2.00,2.00)"
    setup = al.SetupPipeline(subhalo_instance=al.mp.SphericalNFW(centre=(3.0, 4.0)))
    assert setup.subhalo_centre_tag == "__sub_centre_(3.00,4.00)"
    setup = al.SetupPipeline(subhalo_instance=al.mp.SphericalNFW(centre=(3.027, 4.033)))
    assert setup.subhalo_centre_tag == "__sub_centre_(3.03,4.03)"


def test__subhalo_mass_at_200_tag():

    setup = al.SetupPipeline(subhalo_instance=None)
    assert setup.subhalo_mass_at_200_tag == ""
    setup = al.SetupPipeline(
        subhalo_instance=al.mp.SphericalNFWMCRLudlow(mass_at_200=1e8)
    )
    assert setup.subhalo_mass_at_200_tag == "__sub_mass_1.0e+08"
    setup = al.SetupPipeline(
        subhalo_instance=al.mp.SphericalNFWMCRLudlow(mass_at_200=1e9)
    )
    assert setup.subhalo_mass_at_200_tag == "__sub_mass_1.0e+09"
    setup = al.SetupPipeline(
        subhalo_instance=al.mp.SphericalNFWMCRLudlow(mass_at_200=1e10)
    )
    assert setup.subhalo_mass_at_200_tag == "__sub_mass_1.0e+10"


def test__tag():

    setup = al.SetupPipeline(
        hyper_galaxies=True,
        hyper_background_noise=True,
        hyper_image_sky=True,
        align_bulge_dark_centre=True,
    )

    assert (
        setup.tag
        == "setup__hyper_galaxies_bg_sky_bg_noise__with_shear__align_bulge_dark_centre"
    )

    setup = al.SetupPipeline(
        pixelization=al.pix.Rectangular,
        regularization=al.reg.Constant,
        lens_light_centre=(1.0, 2.0),
        lens_mass_centre=(3.0, 4.0),
        align_light_mass_centre=False,
        no_shear=True,
    )

    assert (
        setup.tag
        == "setup__pix_rect__reg_const__lens_light_centre_(1.00,2.00)__no_shear__lens_mass_centre_(3.00,4.00)"
    )

    setup = al.SetupPipeline(align_light_mass_centre=True, number_of_gaussians=1)

    assert setup.tag == "setup__align_light_mass_centre__gaussians_x1__with_shear"

    setup = al.SetupPipeline(
        subhalo_instance=al.mp.SphericalNFWMCRLudlow(centre=(1.0, 2.0), mass_at_200=1e8)
    )

    assert setup.tag == "setup__with_shear__sub_centre_(1.00,2.00)__sub_mass_1.0e+08"

    setup = al.SetupPipeline(include_smbh=True, smbh_centre_fixed=True)

    assert setup.tag == "setup__with_shear__smbh_centre_fixed"


def test__set_mass_to_light_ratios_of_lens():

    lmp_0 = af.PriorModel(al.lmp.EllipticalSersic)
    lmp_1 = af.PriorModel(al.lmp.EllipticalSersic)
    lmp_2 = af.PriorModel(al.lmp.EllipticalSersic)

    setup = al.SetupPipeline(constant_mass_to_light_ratio=False)

    setup.set_mass_to_light_ratios_of_light_and_mass_profiles(
        light_and_mass_profiles=[lmp_0, lmp_1, lmp_2]
    )

    assert lmp_0.mass_to_light_ratio != lmp_1.mass_to_light_ratio
    assert lmp_0.mass_to_light_ratio != lmp_2.mass_to_light_ratio
    assert lmp_1.mass_to_light_ratio != lmp_2.mass_to_light_ratio

    lmp_0 = af.PriorModel(al.lmp.EllipticalSersic)
    lmp_1 = af.PriorModel(al.lmp.EllipticalSersic)
    lmp_2 = af.PriorModel(al.lmp.EllipticalSersic)

    setup = al.SetupPipeline(constant_mass_to_light_ratio=True)

    setup.set_mass_to_light_ratios_of_light_and_mass_profiles(
        light_and_mass_profiles=[lmp_0, lmp_1, lmp_2]
    )

    assert lmp_0.mass_to_light_ratio == lmp_1.mass_to_light_ratio
    assert lmp_0.mass_to_light_ratio == lmp_2.mass_to_light_ratio
    assert lmp_1.mass_to_light_ratio == lmp_2.mass_to_light_ratio


def test__smbh_from_centre():

    setup = al.SetupPipeline(include_smbh=False, smbh_centre_fixed=True)
    smbh = setup.smbh_from_centre(centre=(0.0, 0.0))
    assert smbh is None

    setup = al.SetupPipeline(include_smbh=True, smbh_centre_fixed=True)
    smbh = setup.smbh_from_centre(centre=(0.0, 0.0))
    assert isinstance(smbh, af.PriorModel)
    assert smbh.centre == (0.0, 0.0)

    setup = al.SetupPipeline(include_smbh=True, smbh_centre_fixed=False)
    smbh = setup.smbh_from_centre(centre=(0.1, 0.2), centre_sigma=0.2)
    assert isinstance(smbh, af.PriorModel)
    assert isinstance(smbh.centre[0], af.GaussianPrior)
    assert smbh.centre[0].mean == 0.1
    assert smbh.centre[0].sigma == 0.2
    assert isinstance(smbh.centre[1], af.GaussianPrior)
    assert smbh.centre[1].mean == 0.2
    assert smbh.centre[1].sigma == 0.2



import time

import autolens as al
import numpy as np

repeats = 1

total_visibilities = 500000

real_space_shape_2d = (256, 256)
real_space_pixel_scales = 0.05
real_space_sub_size = 1
real_space_radius = 3.0

image_pixels = real_space_shape_2d[0] * real_space_shape_2d[1]

shape_data = 8 * total_visibilities
shape_preloads = total_visibilities * image_pixels * 2

total_shape = shape_data + shape_preloads

print("Data Memory Use (GB) = " + str(shape_data * 8e-9))
print("PreLoad Memory Use (GB) = " + str(shape_preloads * 8e-9))
print("Total Memory Use (GB) = " + str(total_shape * 8e-9))
print()

# Only delete this if the memory use looks... Okay
# stop

visibilities = al.Visibilities.ones(shape_1d=(total_visibilities,))
uv_wavelengths = np.ones(shape=(total_visibilities, 2))
noise_map = al.VisibilitiesNoiseMap.ones(shape_1d=(total_visibilities,))

interferometer = al.Interferometer(
    visibilities=visibilities, noise_map=noise_map, uv_wavelengths=uv_wavelengths
)

print("Real space sub grid size = " + str(real_space_sub_size))
print("Real space circular mask radius = " + str(real_space_radius) + "\n")

lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mp.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, elliptical_comps=(0.17647, 0.0)
    ),
)

source_galaxy = al.Galaxy(redshift=1.0, light=al.lp.EllipticalSersic(intensity=1.0))

mask = al.Mask.circular(
    shape_2d=real_space_shape_2d,
    pixel_scales=real_space_pixel_scales,
    sub_size=real_space_sub_size,
    radius=real_space_radius,
)

masked_interferometer = al.MaskedInterferometer(
    interferometer=interferometer,
    real_space_mask=mask,
    visibilities_mask=np.full(fill_value=False, shape=visibilities.shape),
    transformer_class=al.TransformerNUFFT,
)

print("Number of points = " + str(masked_interferometer.grid.sub_shape_1d) + "\n")
print(
    "Number of visibilities = "
    + str(masked_interferometer.visibilities.shape_1d)
    + "\n"
)

start_overall = time.time()

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

image = tracer.image_from_grid(grid=masked_interferometer.grid).in_2d_binned[::-1, :]
images = np.repeat(image[:, :, np.newaxis], 100, axis=2)

transformer = al.TransformerNUFFT(
    uv_wavelengths=interferometer.uv_wavelengths,
    grid=masked_interferometer.grid.in_1d_binned.in_radians,
)

start = time.time()
for i in range(repeats):
    for j in range(100):
        transformer.visibilities_from_image(image=image)
diff = time.time() - start
print("Visibilities Time = {}".format(diff / repeats))


transformer = al.TransformerNUFFT(
    uv_wavelengths=interferometer.uv_wavelengths,
    grid=masked_interferometer.grid.in_1d_binned.in_radians,
    batch=100,
)

start = time.time()
for i in range(repeats):
    transformer.visibilities_from_image_batch(image=images)
diff = time.time() - start
print("Visibilities Time = {}".format(diff / repeats))



import autolens as al
import numpy as np


class TestMaskedImaging:
    def test__masked_dataset_via_autoarray(self, imaging_7x7, sub_mask_7x7):

        masked_imaging_7x7 = al.MaskedImaging(imaging=imaging_7x7, mask=sub_mask_7x7)

        assert (masked_imaging_7x7.image.in_1d == np.ones(9)).all()

        assert (
            masked_imaging_7x7.image.in_2d == np.ones((7, 7)) * np.invert(sub_mask_7x7)
        ).all()

        assert (masked_imaging_7x7.noise_map.in_1d == 2.0 * np.ones(9)).all()
        assert (
            masked_imaging_7x7.noise_map.in_2d
            == 2.0 * np.ones((7, 7)) * np.invert(sub_mask_7x7)
        ).all()

        assert (masked_imaging_7x7.psf.in_1d == (1.0 / 9.0) * np.ones(9)).all()
        assert (masked_imaging_7x7.psf.in_2d == (1.0 / 9.0) * np.ones((3, 3))).all()
        assert masked_imaging_7x7.psf.shape_2d == (3, 3)

        assert type(masked_imaging_7x7.convolver) == al.Convolver

    def test__inheritance_from_autoarray(
        self, imaging_7x7, sub_mask_7x7, blurring_grid_7x7
    ):

        masked_imaging_7x7 = al.MaskedImaging(
            imaging=imaging_7x7,
            mask=sub_mask_7x7,
            settings=al.SettingsMaskedImaging(psf_shape_2d=(3, 3)),
        )

        grid = al.Grid.from_mask(mask=sub_mask_7x7)

        assert (masked_imaging_7x7.grid == grid).all()

        blurring_grid = grid.blurring_grid_from_kernel_shape(kernel_shape_2d=(3, 3))

        assert (masked_imaging_7x7.blurring_grid.in_1d == blurring_grid_7x7).all()
        assert (masked_imaging_7x7.blurring_grid == blurring_grid).all()


class TestSimulatorImaging:
    def test__from_tracer_and_grid__same_as_tracer_image(self):
        psf = al.Kernel.from_gaussian(shape_2d=(7, 7), sigma=0.5, pixel_scales=1.0)

        grid = al.Grid.uniform(shape_2d=(20, 20), pixel_scales=0.05, sub_size=1)

        lens_galaxy = al.Galaxy(
            redshift=0.5,
            light=al.lp.EllipticalSersic(intensity=1.0),
            mass=al.mp.EllipticalIsothermal(einstein_radius=1.6),
        )

        source_galaxy = al.Galaxy(
            redshift=1.0, light=al.lp.EllipticalSersic(intensity=0.3)
        )

        tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

        simulator = al.SimulatorImaging(
            psf=psf,
            exposure_time_map=al.Array.full(fill_value=10000.0, shape_2d=grid.shape_2d),
            background_sky_map=al.Array.full(fill_value=100.0, shape_2d=grid.shape_2d),
            add_noise=True,
            noise_seed=1,
        )

        imaging = simulator.from_tracer_and_grid(tracer=tracer, grid=grid)

        assert (imaging.image.in_2d == imaging.image.in_2d).all()
        assert (imaging.psf == imaging.psf).all()
        assert (imaging.noise_map == imaging.noise_map).all()

    def test__from_deflections_and_galaxies__same_as_calculation_using_tracer(self):

        psf = al.Kernel.no_blur(pixel_scales=0.05)

        grid = al.Grid.uniform(shape_2d=(20, 20), pixel_scales=0.05, sub_size=1)

        lens_galaxy = al.Galaxy(
            redshift=0.5, mass=al.mp.EllipticalIsothermal(einstein_radius=1.6)
        )

        source_galaxy = al.Galaxy(
            redshift=1.0, light=al.lp.EllipticalSersic(intensity=0.3)
        )

        tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

        simulator = al.SimulatorImaging(
            psf=psf,
            exposure_time_map=al.Array.full(fill_value=10000.0, shape_2d=grid.shape_2d),
            background_sky_map=al.Array.full(fill_value=100.0, shape_2d=grid.shape_2d),
            add_noise=True,
            noise_seed=1,
        )

        imaging = simulator.from_deflections_and_galaxies(
            deflections=tracer.deflections_from_grid(grid=grid),
            galaxies=[source_galaxy],
        )

        imaging_via_image = simulator.from_image(
            image=tracer.image_from_grid(grid=grid)
        )

        assert (imaging.image.in_2d == imaging_via_image.image.in_2d).all()
        assert (imaging.psf == imaging_via_image.psf).all()
        assert (imaging.noise_map == imaging_via_image.noise_map).all()

    def test__simulate_imaging_from_lens__source_galaxy__compare_to_imaging(self):

        lens_galaxy = al.Galaxy(
            redshift=0.5,
            mass=al.mp.EllipticalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.6, elliptical_comps=(0.17647, 0.0)
            ),
        )

        source_galaxy = al.Galaxy(
            redshift=0.5,
            light=al.lp.EllipticalSersic(
                centre=(0.1, 0.1),
                elliptical_comps=(0.096225, -0.055555),
                intensity=0.3,
                effective_radius=1.0,
                sersic_index=2.5,
            ),
        )

        grid = al.Grid.uniform(shape_2d=(11, 11), pixel_scales=0.2, sub_size=1)

        psf = al.Kernel.no_blur(pixel_scales=0.2)

        simulator = al.SimulatorImaging(
            psf=psf,
            exposure_time_map=al.Array.full(fill_value=10000.0, shape_2d=grid.shape_2d),
            background_sky_map=al.Array.full(fill_value=100.0, shape_2d=grid.shape_2d),
            add_noise=True,
            noise_seed=1,
        )

        imaging = simulator.from_galaxies_and_grid(
            galaxies=[lens_galaxy, source_galaxy], grid=grid
        )

        tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

        imaging_via_image = simulator.from_image(
            image=tracer.image_from_grid(grid=grid)
        )

        assert (imaging.image == imaging_via_image.image).all()
        assert (imaging.psf == imaging_via_image.psf).all()
        assert imaging.noise_map == imaging_via_image.noise_map



from autoarray import preprocess
from autoarray.dataset.imaging import Imaging
from autoarray.dataset.interferometer import Interferometer
from autoarray.mask.mask import Mask
from autoarray.operators.convolver import Convolver
from autoarray.inversion import pixelizations as pix, regularization as reg
from autoarray.inversion.pixelizations import SettingsPixelization
from autoarray.inversion.inversions import inversion as Inversion, SettingsInversion
from autoarray.inversion.mappers import mapper as Mapper
from autoarray.operators.transformer import TransformerDFT
from autoarray.operators.transformer import TransformerNUFFT
from autoarray.operators.transformer import TransformerNUFFT
from autoarray.structures.arrays import Array, Values
from autoarray.structures.grids import (
    Grid,
    GridIterate,
    GridInterpolate,
    GridCoordinates,
    GridRectangular,
    GridVoronoi,
)
from autoarray.structures.kernel import Kernel
from autoarray.structures.visibilities import Visibilities, VisibilitiesNoiseMap
from autogalaxy import dimensions as dim
from autogalaxy import util
from autogalaxy.dataset.imaging import SettingsMaskedImaging
from autogalaxy.dataset.interferometer import SettingsMaskedInterferometer
from autogalaxy.galaxy.fit_galaxy import FitGalaxy
from autogalaxy.galaxy.galaxy import Galaxy, HyperGalaxy, Redshift
from autogalaxy.galaxy.galaxy_data import GalaxyData
from autogalaxy.galaxy.galaxy_model import GalaxyModel
from autogalaxy.hyper import hyper_data
from autogalaxy.pipeline.phase.extensions import CombinedHyperPhase
from autogalaxy.pipeline.phase.extensions import HyperGalaxyPhase
from autogalaxy.pipeline.phase.extensions.hyper_galaxy_phase import HyperGalaxyPhase
from autogalaxy.pipeline.phase.extensions.hyper_phase import HyperPhase
from autogalaxy.pipeline.phase.extensions.inversion_phase import (
    InversionPhase,
    ModelFixingHyperPhase,
)
from autogalaxy.plane.plane import Plane
from autogalaxy.profiles import (
    light_profiles as lp,
    mass_profiles as mp,
    light_and_mass_profiles as lmp,
)
from autogalaxy.util import convert

from . import aggregator as agg
from . import plot
from .dataset.imaging import MaskedImaging, SimulatorImaging
from .dataset.interferometer import MaskedInterferometer, SimulatorInterferometer
from .fit.fit import FitImaging, FitInterferometer
from .fit.fit_positions import FitPositionsSourcePlaneMaxSeparation
from .lens.settings import SettingsLens
from .lens.ray_tracing import Tracer
from .lens.positions_solver import PositionsFinder
from .pipeline.setup import SetupPipeline
from .pipeline import slam
from .pipeline.phase.settings import SettingsPhaseImaging
from .pipeline.phase.settings import SettingsPhaseInterferometer
from .pipeline.phase.imaging.phase import PhaseImaging
from .pipeline.phase.interferometer.phase import PhaseInterferometer
from .pipeline.phase.extensions.stochastic_phase import StochasticPhase
from .pipeline.phase.phase_galaxy import PhaseGalaxy
from .pipeline.pipeline import PipelineDataset, PipelinePositions

__version__ = '1.4.2'



import autofit as af
from astropy import cosmology as cosmo
from autogalaxy.pipeline.phase.interferometer.phase import (
    PhaseAttributes as AgPhaseAttributes,
)
from autolens.dataset import interferometer
from autolens.pipeline.phase import dataset
from autolens.pipeline.phase.settings import SettingsPhaseInterferometer
from autolens.pipeline.phase.interferometer.analysis import Analysis
from autolens.pipeline.phase.interferometer.result import Result


class PhaseInterferometer(dataset.PhaseDataset):
    galaxies = af.PhaseProperty("galaxies")
    hyper_background_noise = af.PhaseProperty("hyper_background_noise")

    Analysis = Analysis
    Result = Result

    @af.convert_paths
    def __init__(
        self,
        paths,
        *,
        search,
        real_space_mask,
        galaxies=None,
        hyper_background_noise=None,
        settings=SettingsPhaseInterferometer(),
        cosmology=cosmo.Planck15,
    ):

        """

        A phase in an lens pipeline. Uses the set non_linear search to try to fit models and hyper_galaxies
        passed to it.

        Parameters
        ----------
        search: class
            The class of a non_linear search
        sub_size: int
            The side length of the subgrid
        """

        paths.tag = settings.phase_tag_with_inversion

        super().__init__(
            paths=paths,
            search=search,
            galaxies=galaxies,
            settings=settings,
            cosmology=cosmology,
        )

        self.hyper_background_noise = hyper_background_noise

        self.is_hyper_phase = False

        self.real_space_mask = real_space_mask

    def make_analysis(self, dataset, mask, results=None):
        """
        Create an lens object. Also calls the prior passing and masked_interferometer modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        positions
        mask: Mask
            The default masks passed in by the pipeline
        dataset: im.Interferometer
            An masked_interferometer that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens : Analysis
            An lens object that the non-linear search calls to determine the fit of a set of values
        """

        masked_interferometer = interferometer.MaskedInterferometer(
            interferometer=dataset,
            visibilities_mask=mask,
            real_space_mask=self.real_space_mask,
            settings=self.settings.settings_masked_interferometer,
        )

        self.output_phase_info()

        analysis = self.Analysis(
            masked_interferometer=masked_interferometer,
            settings=self.settings,
            cosmology=self.cosmology,
            image_path=self.search.paths.image_path,
            results=results,
            log_likelihood_cap=self.settings.log_likelihood_cap,
        )

        return analysis

    def make_phase_attributes(self, analysis):
        return PhaseAttributes(
            cosmology=self.cosmology,
            real_space_mask=self.real_space_mask,
            positions=analysis.masked_dataset.positions,
            hyper_model_image=analysis.hyper_model_image,
            hyper_galaxy_image_path_dict=analysis.hyper_galaxy_image_path_dict,
        )

    def output_phase_info(self):

        file_phase_info = "{}/{}".format(self.search.paths.output_path, "phase.info")

        with open(file_phase_info, "w") as phase_info:
            phase_info.write("Optimizer = {} \n".format(type(self.search).__name__))
            phase_info.write(
                "Sub-grid size = {} \n".format(
                    self.settings.settings_masked_interferometer.sub_size
                )
            )
            phase_info.write(
                "Positions Threshold = {} \n".format(
                    self.settings.settings_lens.positions_threshold
                )
            )
            phase_info.write("Cosmology = {} \n".format(self.cosmology))

            phase_info.close()


class PhaseAttributes(AgPhaseAttributes):
    def __init__(
        self,
        cosmology,
        real_space_mask,
        positions,
        hyper_model_image,
        hyper_galaxy_image_path_dict,
    ):

        super().__init__(
            cosmology=cosmology,
            real_space_mask=real_space_mask,
            hyper_model_image=hyper_model_image,
            hyper_galaxy_image_path_dict=hyper_galaxy_image_path_dict,
        )

        self.positions = positions



import copy
from scipy.stats import norm
import matplotlib.pyplot as plt

from autoconf import conf
import autoarray as aa
from autoarray.plot import mat_objs
from autogalaxy.plot import fit_galaxy_plots, hyper_plots, inversion_plots
from autogalaxy.plot import lensing_plotters
from autolens.plot import ray_tracing_plots, fit_imaging_plots, fit_interferometer_plots


def setting(section, name):
    return conf.instance.visualize_plots.get(section, name, bool)


def plot_setting(section, name):
    return setting(section, name)


class AbstractVisualizer:
    def __init__(self, image_path):

        self.plotter = lensing_plotters.Plotter(
            output=mat_objs.Output(path=image_path, format="png")
        )
        self.sub_plotter = lensing_plotters.SubPlotter(
            output=mat_objs.Output(path=image_path + "subplots/", format="png")
        )
        self.include = lensing_plotters.Include()

        self.plot_ray_tracing_all_at_end_png = plot_setting(
            "ray_tracing", "all_at_end_png"
        )
        self.plot_ray_tracing_all_at_end_fits = plot_setting(
            "ray_tracing", "all_at_end_fits"
        )
        self.plot_subplot_ray_tracing = plot_setting(
            "ray_tracing", "subplot_ray_tracing"
        )
        self.plot_ray_tracing_image = plot_setting("ray_tracing", "image")
        self.plot_ray_tracing_source_plane = plot_setting(
            "ray_tracing", "source_plane_image"
        )
        self.plot_ray_tracing_convergence = plot_setting("ray_tracing", "convergence")
        self.plot_ray_tracing_potential = plot_setting("ray_tracing", "potential")
        self.plot_ray_tracing_deflections = plot_setting("ray_tracing", "deflections")
        self.plot_ray_tracing_magnification = plot_setting(
            "ray_tracing", "magnification"
        )

    def new_visualizer_with_preloaded_critical_curves_and_caustics(
        self, preloaded_critical_curves, preloaded_caustics
    ):

        visualizer = copy.deepcopy(self)

        visualizer.include = visualizer.include.new_include_with_preloaded_critical_curves_and_caustics(
            preloaded_critical_curves=preloaded_critical_curves,
            preloaded_caustics=preloaded_caustics,
        )

        return visualizer


class PhaseGalaxyVisualizer(AbstractVisualizer):
    def __init__(self, image_path):
        super().__init__(image_path)
        self.plot_galaxy_fit_all_at_end_png = plot_setting(
            "galaxy_fit", "all_at_end_png"
        )
        self.plot_galaxy_fit_all_at_end_fits = plot_setting(
            "galaxy_fit", "all_at_end_fits"
        )
        self.plot_subplot_galaxy_fit = plot_setting("galaxy_fit", "subplot_galaxy_fit")
        self.plot_galaxy_fit_image = plot_setting("galaxy_fit", "image")
        self.plot_galaxy_fit_noise_map = plot_setting("galaxy_fit", "noise_map")
        self.plot_galaxy_fit_model_image = plot_setting("galaxy_fit", "model_image")
        self.plot_galaxy_fit_residual_map = plot_setting("galaxy_fit", "residual_map")
        self.plot_galaxy_fit_chi_squared_map = plot_setting(
            "galaxy_fit", "chi_squared_map"
        )

    def plot_galaxy_fit_subplot(self, fit):
        if self.plot_subplot_galaxy_fit:
            fit_galaxy_plots.subplot_fit_galaxy(
                fit=fit, include=self.include, sub_plotter=self.sub_plotter
            )

    def plot_fit_individuals(self, fit):

        fit_galaxy_plots.individuals(
            fit=fit,
            plot_image=self.plot_galaxy_fit_image,
            plot_noise_map=self.plot_galaxy_fit_noise_map,
            plot_model_image=self.plot_galaxy_fit_model_image,
            plot_residual_map=self.plot_galaxy_fit_residual_map,
            plot_chi_squared_map=self.plot_galaxy_fit_chi_squared_map,
            include=self.include,
            plotter=self.plotter,
        )


class PhaseDatasetVisualizer(AbstractVisualizer):
    def __init__(self, masked_dataset, image_path):
        super().__init__(image_path)
        self.masked_dataset = masked_dataset

        self.plot_subplot_dataset = plot_setting("dataset", "subplot_dataset")
        self.plot_dataset_data = plot_setting("dataset", "data")
        self.plot_dataset_noise_map = plot_setting("dataset", "noise_map")
        self.plot_dataset_psf = plot_setting("dataset", "psf")
        self.plot_dataset_inverse_noise_map = plot_setting(
            "dataset", "inverse_noise_map"
        )
        self.plot_dataset_signal_to_noise_map = plot_setting(
            "dataset", "signal_to_noise_map"
        )
        self.plot_dataset_absolute_signal_to_noise_map = plot_setting(
            "dataset", "absolute_signal_to_noise_map"
        )
        self.plot_dataset_potential_chi_squared_map = plot_setting(
            "dataset", "potential_chi_squared_map"
        )

        self.plot_fit_all_at_end_png = plot_setting("fit", "all_at_end_png")
        self.plot_fit_all_at_end_fits = plot_setting("fit", "all_at_end_fits")
        self.plot_subplot_fit = plot_setting("fit", "subplot_fit")
        self.plot_subplots_of_all_planes_fits = plot_setting(
            "fit", "subplots_of_plane_fits"
        )

        self.plot_fit_data = plot_setting("fit", "data")
        self.plot_fit_noise_map = plot_setting("fit", "noise_map")
        self.plot_fit_signal_to_noise_map = plot_setting("fit", "signal_to_noise_map")
        self.plot_fit_model_data = plot_setting("fit", "model_data")
        self.plot_fit_residual_map = plot_setting("fit", "residual_map")
        self.plot_fit_normalized_residual_map = plot_setting(
            "fit", "normalized_residual_map"
        )
        self.plot_fit_chi_squared_map = plot_setting("fit", "chi_squared_map")
        self.plot_fit_subtracted_images_of_planes = plot_setting(
            "fit", "subtracted_images_of_planes"
        )
        self.plot_fit_model_images_of_planes = plot_setting(
            "fit", "model_images_of_planes"
        )
        self.plot_fit_plane_images_of_planes = plot_setting(
            "fit", "plane_images_of_planes"
        )

        self.plot_subplot_inversion = plot_setting("inversion", "subplot_inversion")
        self.plot_inversion_reconstructed_image = plot_setting(
            "inversion", "reconstructed_image"
        )

        self.plot_inversion_reconstruction = plot_setting("inversion", "reconstruction")

        self.plot_inversion_errors = plot_setting("inversion", "errors")

        self.plot_inversion_residual_map = plot_setting("inversion", "residual_map")
        self.plot_inversion_normalized_residual_map = plot_setting(
            "inversion", "normalized_residual_map"
        )
        self.plot_inversion_chi_squared_map = plot_setting(
            "inversion", "chi_squared_map"
        )
        self.plot_inversion_regularization_weights = plot_setting(
            "inversion", "regularization_weight_map"
        )
        self.plot_inversion_interpolated_reconstruction = plot_setting(
            "inversion", "interpolated_reconstruction"
        )
        self.plot_inversion_interpolated_errors = plot_setting(
            "inversion", "interpolated_errors"
        )

        self.plot_hyper_model_image = plot_setting("hyper_galaxy", "model_image")
        self.plot_hyper_galaxy_images = plot_setting("hyper_galaxy", "images")

        self.plot_stochastic_histogram = plot_setting("other", "stochastic_histogram")

    def visualize_ray_tracing(self, tracer, during_analysis):

        plotter = self.plotter.plotter_with_new_output(
            path=self.plotter.output.path + "ray_tracing/"
        )

        if self.plot_subplot_ray_tracing:

            ray_tracing_plots.subplot_tracer(
                tracer=tracer,
                grid=self.masked_dataset.grid,
                positions=self.include.positions_from_masked_dataset(
                    masked_dataset=self.masked_dataset
                ),
                include=self.include,
                sub_plotter=self.sub_plotter,
            )

        ray_tracing_plots.individual(
            tracer=tracer,
            grid=self.masked_dataset.grid,
            positions=self.include.positions_from_masked_dataset(
                masked_dataset=self.masked_dataset
            ),
            plot_image=self.plot_ray_tracing_image,
            plot_source_plane=self.plot_ray_tracing_source_plane,
            plot_convergence=self.plot_ray_tracing_convergence,
            plot_potential=self.plot_ray_tracing_potential,
            plot_deflections=self.plot_ray_tracing_deflections,
            plot_magnification=self.plot_ray_tracing_magnification,
            include=self.include,
            plotter=plotter,
        )

        if not during_analysis:

            if self.plot_ray_tracing_all_at_end_png:

                ray_tracing_plots.individual(
                    tracer=tracer,
                    grid=self.masked_dataset.grid,
                    positions=self.include.positions_from_masked_dataset(
                        masked_dataset=self.masked_dataset
                    ),
                    plot_image=True,
                    plot_source_plane=True,
                    plot_convergence=True,
                    plot_potential=True,
                    plot_deflections=True,
                    include=self.include,
                    plotter=plotter,
                )

            if self.plot_ray_tracing_all_at_end_fits:

                self.visualize_ray_tracing_in_fits(tracer=tracer)

    def visualize_ray_tracing_in_fits(self, tracer):

        fits_plotter = self.plotter.plotter_with_new_output(
            path=self.plotter.output.path + "/ray_tracing/fits/", format="fits"
        )

        ray_tracing_plots.individual(
            tracer=tracer,
            grid=self.masked_dataset.grid,
            plot_image=True,
            plot_source_plane=True,
            plot_convergence=True,
            plot_potential=True,
            plot_deflections=True,
            include=self.include,
            plotter=fits_plotter,
        )

    def visualize_hyper_images(self, hyper_galaxy_image_path_dict, hyper_model_image):

        plotter = self.plotter.plotter_with_new_output(
            path=self.plotter.output.path + "hyper/"
        )

        sub_plotter = self.sub_plotter.plotter_with_new_output(
            path=self.plotter.output.path + "hyper/"
        )

        if self.plot_hyper_model_image:
            hyper_plots.hyper_model_image(
                hyper_model_image=hyper_model_image,
                mask=self.include.mask_from_masked_dataset(
                    masked_dataset=self.masked_dataset
                ),
                include=self.include,
                plotter=plotter,
            )

        if self.plot_hyper_galaxy_images:
            hyper_plots.subplot_hyper_galaxy_images(
                hyper_galaxy_image_path_dict=hyper_galaxy_image_path_dict,
                mask=self.include.mask_from_masked_dataset(
                    masked_dataset=self.masked_dataset
                ),
                include=self.include,
                sub_plotter=sub_plotter,
            )

    def visualize_stochastic_histogram(
        self, log_evidences, max_log_evidence, during_analysis, histogram_bins=10
    ):

        if log_evidences is None:
            return

        plotter = self.plotter.plotter_with_new_output(
            path=self.plotter.output.path + "other/"
        )

        if self.plot_stochastic_histogram and not during_analysis:

            (mu, sigma) = norm.fit(log_evidences)
            n, bins, patches = plt.hist(x=log_evidences, bins=histogram_bins, density=1)
            y = norm.pdf(bins, mu, sigma)
            plt.plot(bins, y, "--")
            plt.xlabel("log evidence")
            plt.title("Stochastic Log Evidence Histogram")
            plt.axvline(max_log_evidence, color="r")
            plt.savefig(
                f"{plotter.output.path}stochastic_histogram.png", bbox_inches="tight"
            )


class PhaseImagingVisualizer(PhaseDatasetVisualizer):
    def __init__(self, masked_dataset, image_path, results=None):
        super(PhaseImagingVisualizer, self).__init__(
            masked_dataset=masked_dataset, image_path=image_path
        )

        self.plot_dataset_psf = plot_setting("dataset", "psf")

        self.plot_hyper_model_image = plot_setting("hyper_galaxy", "model_image")
        self.plot_hyper_galaxy_images = plot_setting("hyper_galaxy", "images")

        self.visualize_imaging()

    @property
    def masked_imaging(self):
        return self.masked_dataset

    def visualize_imaging(self):

        plotter = self.plotter.plotter_with_new_output(
            path=self.plotter.output.path + "imaging/"
        )

        if self.plot_subplot_dataset:
            aa.plot.Imaging.subplot_imaging(
                imaging=self.masked_imaging.imaging,
                mask=self.include.mask_from_masked_dataset(
                    masked_dataset=self.masked_dataset
                ),
                positions=self.include.positions_from_masked_dataset(
                    masked_dataset=self.masked_dataset
                ),
                include=self.include,
                sub_plotter=self.sub_plotter,
            )

        aa.plot.Imaging.individual(
            imaging=self.masked_imaging.imaging,
            mask=self.include.mask_from_masked_dataset(
                masked_dataset=self.masked_dataset
            ),
            positions=self.include.positions_from_masked_dataset(
                masked_dataset=self.masked_dataset
            ),
            plot_image=self.plot_dataset_data,
            plot_noise_map=self.plot_dataset_noise_map,
            plot_psf=self.plot_dataset_psf,
            plot_inverse_noise_map=self.plot_dataset_inverse_noise_map,
            plot_signal_to_noise_map=self.plot_dataset_signal_to_noise_map,
            plot_absolute_signal_to_noise_map=self.plot_dataset_absolute_signal_to_noise_map,
            plot_potential_chi_squared_map=self.plot_dataset_potential_chi_squared_map,
            include=self.include,
            plotter=plotter,
        )

    def visualize_fit(self, fit, during_analysis):

        plotter = self.plotter.plotter_with_new_output(
            path=self.plotter.output.path + "fit_imaging/"
        )

        if self.plot_subplot_fit:
            fit_imaging_plots.subplot_fit_imaging(
                fit=fit, include=self.include, sub_plotter=self.sub_plotter
            )

        if self.plot_subplots_of_all_planes_fits:
            fit_imaging_plots.subplots_of_all_planes(
                fit=fit, include=self.include, sub_plotter=self.sub_plotter
            )

        fit_imaging_plots.individuals(
            fit=fit,
            plot_image=self.plot_fit_data,
            plot_noise_map=self.plot_fit_noise_map,
            plot_signal_to_noise_map=self.plot_fit_signal_to_noise_map,
            plot_model_image=self.plot_fit_model_data,
            plot_residual_map=self.plot_fit_residual_map,
            plot_chi_squared_map=self.plot_fit_chi_squared_map,
            plot_normalized_residual_map=self.plot_fit_normalized_residual_map,
            plot_subtracted_images_of_planes=self.plot_fit_subtracted_images_of_planes,
            plot_model_images_of_planes=self.plot_fit_model_images_of_planes,
            plot_plane_images_of_planes=self.plot_fit_plane_images_of_planes,
            include=self.include,
            plotter=plotter,
        )

        if fit.inversion is not None:

            if self.plot_subplot_inversion:
                inversion_plots.subplot_inversion(
                    inversion=fit.inversion,
                    image_positions=self.include.positions_from_fit(fit=fit),
                    source_positions=self.include.positions_of_plane_from_fit_and_plane_index(
                        fit=fit, plane_index=-1
                    ),
                    grid=self.include.inversion_image_pixelization_grid_from_fit(
                        fit=fit
                    ),
                    light_profile_centres=self.include.light_profile_centres_from_obj(
                        obj=fit.tracer.image_plane
                    ),
                    mass_profile_centres=self.include.mass_profile_centres_from_obj(
                        obj=fit.tracer.image_plane
                    ),
                    critical_curves=self.include.critical_curves_from_obj(
                        obj=fit.tracer
                    ),
                    caustics=self.include.caustics_from_obj(obj=fit.tracer),
                    include=self.include,
                    sub_plotter=self.sub_plotter,
                )

            plotter = self.plotter.plotter_with_new_output(
                path=self.plotter.output.path + "inversion/"
            )

            inversion_plots.individuals(
                inversion=fit.inversion,
                image_positions=self.include.positions_from_fit(fit=fit),
                source_positions=self.include.positions_of_plane_from_fit_and_plane_index(
                    fit=fit, plane_index=-1
                ),
                grid=self.include.inversion_image_pixelization_grid_from_fit(fit=fit),
                light_profile_centres=self.include.light_profile_centres_from_obj(
                    obj=fit.tracer.image_plane
                ),
                mass_profile_centres=self.include.mass_profile_centres_from_obj(
                    obj=fit.tracer.image_plane
                ),
                critical_curves=self.include.critical_curves_from_obj(obj=fit.tracer),
                caustics=self.include.caustics_from_obj(obj=fit.tracer),
                plot_reconstructed_image=self.plot_inversion_reconstruction,
                plot_reconstruction=self.plot_inversion_reconstruction,
                plot_errors=self.plot_inversion_errors,
                plot_residual_map=self.plot_inversion_residual_map,
                plot_normalized_residual_map=self.plot_inversion_normalized_residual_map,
                plot_chi_squared_map=self.plot_inversion_chi_squared_map,
                plot_regularization_weight_map=self.plot_inversion_regularization_weights,
                plot_interpolated_reconstruction=self.plot_inversion_interpolated_reconstruction,
                plot_interpolated_errors=self.plot_inversion_interpolated_errors,
                include=self.include,
                plotter=plotter,
            )

        if not during_analysis:

            if self.plot_fit_all_at_end_png:
                fit_imaging_plots.individuals(
                    fit=fit,
                    plot_image=True,
                    plot_noise_map=True,
                    plot_signal_to_noise_map=True,
                    plot_model_image=True,
                    plot_residual_map=True,
                    plot_normalized_residual_map=True,
                    plot_chi_squared_map=True,
                    plot_subtracted_images_of_planes=True,
                    plot_model_images_of_planes=True,
                    plot_plane_images_of_planes=True,
                    include=self.include,
                    plotter=plotter,
                )

                if fit.inversion is not None:
                    inversion_plots.individuals(
                        inversion=fit.inversion,
                        image_positions=self.include.positions_from_fit(fit=fit),
                        source_positions=self.include.positions_of_plane_from_fit_and_plane_index(
                            fit=fit, plane_index=-1
                        ),
                        grid=self.include.inversion_image_pixelization_grid_from_fit(
                            fit=fit
                        ),
                        light_profile_centres=self.include.light_profile_centres_from_obj(
                            obj=fit.tracer.image_plane
                        ),
                        mass_profile_centres=self.include.mass_profile_centres_from_obj(
                            obj=fit.tracer.image_plane
                        ),
                        critical_curves=self.include.critical_curves_from_obj(
                            obj=fit.tracer
                        ),
                        caustics=self.include.caustics_from_obj(obj=fit.tracer),
                        plot_reconstructed_image=True,
                        plot_reconstruction=True,
                        plot_errors=True,
                        plot_residual_map=True,
                        plot_normalized_residual_map=True,
                        plot_chi_squared_map=True,
                        plot_regularization_weight_map=True,
                        plot_interpolated_reconstruction=True,
                        plot_interpolated_errors=True,
                        include=self.include,
                        plotter=plotter,
                    )

            if self.plot_fit_all_at_end_fits:

                self.visualize_fit_in_fits(fit=fit)

    def visualize_fit_in_fits(self, fit):

        fits_plotter = self.plotter.plotter_with_new_output(
            path=self.plotter.output.path + "fit_imaging/fits/", format="fits"
        )

        fit_imaging_plots.individuals(
            fit=fit,
            plot_image=True,
            plot_noise_map=True,
            plot_signal_to_noise_map=True,
            plot_model_image=True,
            plot_residual_map=True,
            plot_normalized_residual_map=True,
            plot_chi_squared_map=True,
            plot_subtracted_images_of_planes=True,
            plot_model_images_of_planes=True,
            plot_plane_images_of_planes=True,
            include=self.include,
            plotter=fits_plotter,
        )

        if fit.inversion is not None:

            fits_plotter = self.plotter.plotter_with_new_output(
                path=self.plotter.output.path + "inversion/fits/", format="fits"
            )

            inversion_plots.individuals(
                inversion=fit.inversion,
                plot_reconstructed_image=True,
                plot_interpolated_reconstruction=True,
                plot_interpolated_errors=True,
                include=self.include,
                plotter=fits_plotter,
            )


class PhaseInterferometerVisualizer(PhaseDatasetVisualizer):
    def __init__(self, masked_dataset, image_path):
        super(PhaseInterferometerVisualizer, self).__init__(
            masked_dataset=masked_dataset, image_path=image_path
        )

        self.plot_dataset_uv_wavelengths = plot_setting("dataset", "uv_wavelengths")

        self.plot_hyper_model_image = plot_setting("hyper_galaxy", "model_image")
        self.plot_hyper_galaxy_images = plot_setting("hyper_galaxy", "images")

        self.visualize_interferometer()

    @property
    def masked_interferometer(self):
        return self.masked_dataset

    def visualize_interferometer(self):

        plotter = self.plotter.plotter_with_new_output(
            path=self.plotter.output.path + "interferometer/"
        )

        if self.plot_subplot_dataset:
            aa.plot.Interferometer.subplot_interferometer(
                interferometer=self.masked_dataset.interferometer,
                include=self.include,
                sub_plotter=self.sub_plotter,
            )

        aa.plot.Interferometer.individual(
            interferometer=self.masked_dataset.interferometer,
            plot_visibilities=self.plot_dataset_data,
            plot_u_wavelengths=self.plot_dataset_uv_wavelengths,
            plot_v_wavelengths=self.plot_dataset_uv_wavelengths,
            include=self.include,
            plotter=plotter,
        )

    def visualize_fit(self, fit, during_analysis):

        plotter = self.plotter.plotter_with_new_output(
            path=self.plotter.output.path + "fit_interferometer/"
        )

        if self.plot_subplot_fit:
            fit_interferometer_plots.subplot_fit_interferometer(
                fit=fit, include=self.include, sub_plotter=self.sub_plotter
            )

            fit_interferometer_plots.subplot_fit_real_space(
                fit=fit, include=self.include, sub_plotter=self.sub_plotter
            )

        fit_interferometer_plots.individuals(
            fit=fit,
            plot_visibilities=self.plot_fit_data,
            plot_noise_map=self.plot_fit_noise_map,
            plot_signal_to_noise_map=self.plot_fit_signal_to_noise_map,
            plot_model_visibilities=self.plot_fit_model_data,
            plot_residual_map=self.plot_fit_residual_map,
            plot_chi_squared_map=self.plot_fit_chi_squared_map,
            plot_normalized_residual_map=self.plot_fit_normalized_residual_map,
            include=self.include,
            plotter=plotter,
        )

        if fit.inversion is not None:

            plotter = self.plotter.plotter_with_new_output(
                path=self.plotter.output.path + "inversion/"
            )

            # if self.plot_fit_inversion_as_subplot:
            #     inversion_plots.subplot_inversion(
            #         inversion=fit.inversion,
            #         image_positions=self.include.positions_from_fit(fit=fit),
            #         source_positions=self.include.positions_of_plane_from_fit_and_plane_index(
            #             fit=fit, plane_index=-1
            #         ),
            #         grid=self.include.inversion_image_pixelization_grid_from_fit(fit=fit),
            #         light_profile_centres=self.include.light_profile_centres_of_planes_from_obj(
            #             obj=fit.tracer.image_plane
            #         ),
            #         mass_profile_centres=self.include.mass_profile_centres_of_planes_from_obj(
            #             obj=fit.tracer.image_plane
            #         ),
            #         critical_curves=self.include.critical_curves_from_obj(obj=fit.tracer),
            #         caustics=self.include.caustics_from_obj(obj=fit.tracer),
            #         include=self.include,
            #         sub_plotter=self.sub_plotter,
            #     )

            inversion_plots.individuals(
                inversion=fit.inversion,
                image_positions=self.include.positions_from_fit(fit=fit),
                source_positions=self.include.positions_of_plane_from_fit_and_plane_index(
                    fit=fit, plane_index=-1
                ),
                grid=self.include.inversion_image_pixelization_grid_from_fit(fit=fit),
                light_profile_centres=self.include.light_profile_centres_from_obj(
                    obj=fit.tracer.image_plane
                ),
                mass_profile_centres=self.include.mass_profile_centres_from_obj(
                    obj=fit.tracer.image_plane
                ),
                critical_curves=self.include.critical_curves_from_obj(obj=fit.tracer),
                caustics=self.include.caustics_from_obj(obj=fit.tracer),
                plot_reconstructed_image=self.plot_inversion_reconstruction,
                plot_reconstruction=self.plot_inversion_reconstruction,
                plot_errors=self.plot_inversion_errors,
                #   plot_residual_map=self.plot_fit_inversion_residual_map,
                #   plot_normalized_residual_map=self.plot_fit_inversion_normalized_residual_map,
                #   plot_chi_squared_map=self.plot_fit_inversion_chi_squared_map,
                plot_regularization_weight_map=self.plot_inversion_regularization_weights,
                plot_interpolated_reconstruction=self.plot_inversion_interpolated_reconstruction,
                plot_interpolated_errors=self.plot_inversion_interpolated_errors,
                include=self.include,
                plotter=plotter,
            )

        if not during_analysis:

            if self.plot_fit_all_at_end_png:
                fit_interferometer_plots.individuals(
                    fit=fit,
                    plot_visibilities=True,
                    plot_noise_map=True,
                    plot_signal_to_noise_map=True,
                    plot_model_visibilities=True,
                    plot_residual_map=True,
                    plot_normalized_residual_map=True,
                    plot_chi_squared_map=True,
                    include=self.include,
                    plotter=plotter,
                )

                if fit.inversion is not None:
                    inversion_plots.individuals(
                        inversion=fit.inversion,
                        image_positions=self.include.positions_from_fit(fit=fit),
                        source_positions=self.include.positions_of_plane_from_fit_and_plane_index(
                            fit=fit, plane_index=-1
                        ),
                        grid=self.include.inversion_image_pixelization_grid_from_fit(
                            fit=fit
                        ),
                        light_profile_centres=self.include.light_profile_centres_from_obj(
                            obj=fit.tracer.image_plane
                        ),
                        mass_profile_centres=self.include.mass_profile_centres_from_obj(
                            obj=fit.tracer.image_plane
                        ),
                        critical_curves=self.include.critical_curves_from_obj(
                            obj=fit.tracer
                        ),
                        caustics=self.include.caustics_from_obj(obj=fit.tracer),
                        plot_reconstructed_image=True,
                        plot_reconstruction=True,
                        plot_errors=True,
                        #     plot_residual_map=True,
                        #     plot_normalized_residual_map=True,
                        #     plot_chi_squared_map=True,
                        plot_regularization_weight_map=True,
                        plot_interpolated_reconstruction=True,
                        plot_interpolated_errors=True,
                        include=self.include,
                        plotter=plotter,
                    )

            if self.plot_fit_all_at_end_fits:
                fits_plotter = plotter.plotter_with_new_output(
                    path=plotter.output.path + "/fits/", format="fits"
                )

                fit_interferometer_plots.individuals(
                    fit=fit,
                    plot_visibilities=True,
                    plot_noise_map=True,
                    plot_signal_to_noise_map=True,
                    plot_model_visibilities=True,
                    plot_residual_map=True,
                    plot_normalized_residual_map=True,
                    plot_chi_squared_map=True,
                    include=self.include,
                    plotter=fits_plotter,
                )

                if fit.inversion is not None:
                    inversion_plots.individuals(
                        inversion=fit.inversion,
                        image_positions=self.include.positions_from_fit(fit=fit),
                        source_positions=self.include.positions_of_plane_from_fit_and_plane_index(
                            fit=fit, plane_index=-1
                        ),
                        grid=self.include.inversion_image_pixelization_grid_from_fit(
                            fit=fit
                        ),
                        light_profile_centres=self.include.light_profile_centres_from_obj(
                            obj=fit.tracer.image_plane
                        ),
                        mass_profile_centres=self.include.mass_profile_centres_from_obj(
                            obj=fit.tracer.image_plane
                        ),
                        critical_curves=self.include.critical_curves_from_obj(
                            obj=fit.tracer
                        ),
                        caustics=self.include.caustics_from_obj(obj=fit.tracer),
                        plot_reconstructed_image=True,
                        plot_interpolated_reconstruction=True,
                        plot_interpolated_errors=True,
                        include=self.include,
                        plotter=plotter,
                    )


class HyperGalaxyVisualizer(AbstractVisualizer):
    def __init__(self, image_path):
        super().__init__(image_path)
        self.plot_hyper_galaxy_subplot = plot_setting(
            "hyper_galaxy", "subplot_hyper_galaxy"
        )

    def visualize_hyper_galaxy(self, fit, hyper_fit, galaxy_image, contribution_map_in):
        hyper_plots.subplot_fit_hyper_galaxy(
            fit=fit,
            hyper_fit=hyper_fit,
            galaxy_image=galaxy_image,
            contribution_map_in=contribution_map_in,
            include=self.include,
            sub_plotter=self.sub_plotter,
        )



import numpy as np
from autoarray.plot import plotters
from autogalaxy.plot import lensing_plotters, plane_plots, inversion_plots


@lensing_plotters.set_include_and_sub_plotter
@plotters.set_subplot_filename
def subplot_fit_imaging(fit, include=None, sub_plotter=None):
    number_subplots = 6

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=1)

    image(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=2)

    signal_to_noise_map(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=3)

    model_image(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=4)

    residual_map(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=5)

    normalized_residual_map(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=6)

    chi_squared_map(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


def subplots_of_all_planes(fit, include=None, sub_plotter=None):

    for plane_index in range(fit.tracer.total_planes):

        if (
            fit.tracer.planes[plane_index].has_light_profile
            or fit.tracer.planes[plane_index].has_pixelization
        ):

            subplot_of_plane(
                fit=fit,
                plane_index=plane_index,
                include=include,
                sub_plotter=sub_plotter,
            )


@lensing_plotters.set_include_and_sub_plotter
@plotters.set_subplot_filename
def subplot_of_plane(fit, plane_index, include=None, sub_plotter=None):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    number_subplots = 4

    sub_plotter = sub_plotter.plotter_with_new_output(
        filename=sub_plotter.output.filename + "_" + str(plane_index)
    )

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=1)

    image(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=2)

    subtracted_image_of_plane(
        fit=fit, plane_index=plane_index, include=include, plotter=sub_plotter
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=3)

    model_image_of_plane(
        fit=fit, plane_index=plane_index, include=include, plotter=sub_plotter
    )

    if not fit.tracer.planes[plane_index].has_pixelization:

        sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=4)

        traced_grids = fit.tracer.traced_grids_of_planes_from_grid(grid=fit.grid)

        plane_plots.plane_image(
            plane=fit.tracer.planes[plane_index],
            grid=traced_grids[plane_index],
            positions=include.positions_of_plane_from_fit_and_plane_index(
                fit=fit, plane_index=plane_index
            ),
            caustics=include.caustics_from_obj(obj=fit.tracer),
            include=include,
            plotter=sub_plotter,
        )

    elif fit.tracer.planes[plane_index].has_pixelization:

        ratio = float(
            (
                fit.inversion.mapper.grid.scaled_maxima[1]
                - fit.inversion.mapper.grid.scaled_minima[1]
            )
            / (
                fit.inversion.mapper.grid.scaled_maxima[0]
                - fit.inversion.mapper.grid.scaled_minima[0]
            )
        )

        if sub_plotter.figure.aspect in "square":
            aspect_inv = ratio
        elif sub_plotter.figure.aspect in "auto":
            aspect_inv = 1.0 / ratio
        elif sub_plotter.figure.aspect in "equal":
            aspect_inv = 1.0

        sub_plotter.setup_subplot(
            number_subplots=number_subplots, subplot_index=4, aspect=float(aspect_inv)
        )

        inversion_plots.reconstruction(
            inversion=fit.inversion,
            source_positions=include.positions_of_plane_from_fit_and_plane_index(
                fit=fit, plane_index=plane_index
            ),
            caustics=include.caustics_from_obj(obj=fit.tracer),
            include=include,
            plotter=sub_plotter,
        )

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


def individuals(
    fit,
    plot_image=False,
    plot_noise_map=False,
    plot_signal_to_noise_map=False,
    plot_model_image=False,
    plot_residual_map=False,
    plot_normalized_residual_map=False,
    plot_chi_squared_map=False,
    plot_subtracted_images_of_planes=False,
    plot_model_images_of_planes=False,
    plot_plane_images_of_planes=False,
    include=None,
    plotter=None,
):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    if plot_image:

        image(fit=fit, include=include, plotter=plotter)

    if plot_noise_map:

        noise_map(fit=fit, include=include, plotter=plotter)

    if plot_signal_to_noise_map:

        signal_to_noise_map(fit=fit, include=include, plotter=plotter)

    if plot_model_image:

        model_image(fit=fit, include=include, plotter=plotter)

    if plot_residual_map:

        residual_map(fit=fit, include=include, plotter=plotter)

    if plot_normalized_residual_map:

        normalized_residual_map(fit=fit, include=include, plotter=plotter)

    if plot_chi_squared_map:

        chi_squared_map(fit=fit, include=include, plotter=plotter)

    if plot_subtracted_images_of_planes:

        for plane_index in range(fit.tracer.total_planes):

            subtracted_image_of_plane(
                fit=fit, plane_index=plane_index, include=include, plotter=plotter
            )

    if plot_model_images_of_planes:

        for plane_index in range(fit.tracer.total_planes):

            model_image_of_plane(
                fit=fit, plane_index=plane_index, include=include, plotter=plotter
            )

    if plot_plane_images_of_planes:

        for plane_index in range(fit.tracer.total_planes):

            plotter = plotter.plotter_with_new_output(
                filename="plane_image_of_plane_" + str(plane_index)
            )

            if fit.tracer.planes[plane_index].has_light_profile:

                plane_plots.plane_image(
                    plane=fit.tracer.planes[plane_index],
                    grid=include.traced_grid_of_plane_from_fit_and_plane_index(
                        fit=fit, plane_index=plane_index
                    ),
                    positions=include.positions_of_plane_from_fit_and_plane_index(
                        fit=fit, plane_index=plane_index
                    ),
                    caustics=include.caustics_from_obj(obj=fit.tracer),
                    include=include,
                    plotter=plotter,
                )

            elif fit.tracer.planes[plane_index].has_pixelization:

                inversion_plots.reconstruction(
                    inversion=fit.inversion,
                    source_positions=include.positions_of_plane_from_fit_and_plane_index(
                        fit=fit, plane_index=plane_index
                    ),
                    caustics=include.caustics_from_obj(obj=fit.tracer),
                    include=include,
                    plotter=plotter,
                )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def subtracted_image_of_plane(fit, plane_index, include=None, plotter=None):
    """Plot the model image of a specific plane of a lens fit.

    Set *autolens.datas.arrays.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which includes a list of every model image, residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the model image is plotted.
    plane_indexes : int
        The plane from which the model image is generated.
    """

    if isinstance(plotter, lensing_plotters.Plotter):
        plotter = plotter.plotter_with_new_output(
            filename=plotter.output.filename + "_" + str(plane_index)
        )

    if fit.tracer.total_planes > 1:

        other_planes_model_images = [
            model_image
            for i, model_image in enumerate(fit.model_images_of_planes)
            if i != plane_index
        ]

        subtracted_image = fit.image - sum(other_planes_model_images)

    else:

        subtracted_image = fit.image

    plotter_norm = plotter.plotter_with_new_cmap(
        norm_max=np.max(fit.model_images_of_planes[plane_index]),
        norm_min=np.min(fit.model_images_of_planes[plane_index]),
    )

    plotter_norm.plot_array(
        array=subtracted_image,
        mask=include.mask_from_fit(fit=fit),
        grid=include.inversion_image_pixelization_grid_from_fit(fit=fit),
        positions=include.positions_from_fit(fit=fit),
        critical_curves=include.critical_curves_from_obj(obj=fit.tracer),
        light_profile_centres=include.light_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        mass_profile_centres=include.mass_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def model_image_of_plane(fit, plane_index, include=None, plotter=None):
    """Plot the model image of a specific plane of a lens fit.

    Set *autolens.datas.arrays.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which includes a list of every model image, residual_map, chi-squareds, etc.
    plane_indexes : [int]
        The plane from which the model image is generated.
    """

    if isinstance(plotter, lensing_plotters.Plotter):
        plotter = plotter.plotter_with_new_output(
            filename=plotter.output.filename + "_" + str(plane_index)
        )

    plotter.plot_array(
        array=fit.model_images_of_planes[plane_index],
        mask=include.mask_from_fit(fit=fit),
        positions=include.positions_from_fit(fit=fit),
        critical_curves=include.critical_curves_from_obj(obj=fit.tracer),
        light_profile_centres=include.light_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        mass_profile_centres=include.mass_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def image(fit, include=None, plotter=None):
    """Plot the image of a lens fit.

    Set *autolens.datas.array.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : datas.imaging.datas.Imaging
        The datas-datas, which include the observed datas, noise_map, PSF, signal-to-noise_map, etc.
    origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """

    plotter.plot_array(
        array=fit.data,
        mask=include.mask_from_fit(fit=fit),
        grid=include.inversion_image_pixelization_grid_from_fit(fit=fit),
        positions=include.positions_from_fit(fit=fit),
        light_profile_centres=include.light_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        mass_profile_centres=include.mass_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        critical_curves=include.critical_curves_from_obj(obj=fit.tracer),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def noise_map(fit, include=None, plotter=None):
    """Plot the noise-map of a lens fit.

    Set *autolens.datas.array.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : datas.imaging.datas.Imaging
        The datas-datas, which include the observed datas, noise_map, PSF, signal-to-noise_map, etc.
    origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """
    plotter.plot_array(
        array=fit.noise_map,
        mask=include.mask_from_fit(fit=fit),
        light_profile_centres=include.light_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        mass_profile_centres=include.mass_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        critical_curves=include.critical_curves_from_obj(obj=fit.tracer),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def signal_to_noise_map(fit, include=None, plotter=None):
    """Plot the noise-map of a lens fit.

    Set *autolens.datas.array.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : datas.imaging.datas.Imaging
    The datas-datas, which include the observed datas, signal_to_noise_map, PSF, signal-to-signal_to_noise_map, etc.
    origin : True
    If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """
    plotter.plot_array(
        array=fit.signal_to_noise_map,
        mask=include.mask_from_fit(fit=fit),
        light_profile_centres=include.light_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        mass_profile_centres=include.mass_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        critical_curves=include.critical_curves_from_obj(obj=fit.tracer),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def model_image(fit, include=None, plotter=None):
    """Plot the model image of a fit.

    Set *autolens.datas.array.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which include a list of every model image, residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the model image is plotted.
    """
    plotter.plot_array(
        array=fit.model_data,
        mask=include.mask_from_fit(fit=fit),
        positions=include.positions_from_fit(fit=fit),
        light_profile_centres=include.light_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        mass_profile_centres=include.mass_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        critical_curves=include.critical_curves_from_obj(obj=fit.tracer),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def residual_map(fit, include=None, plotter=None):
    """Plot the residual-map of a lens fit.

    Set *autolens.datas.array.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which include a list of every model image, residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the residual_map are plotted.
    """
    plotter.plot_array(
        array=fit.residual_map,
        mask=include.mask_from_fit(fit=fit),
        light_profile_centres=include.light_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        mass_profile_centres=include.mass_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        critical_curves=include.critical_curves_from_obj(obj=fit.tracer),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def normalized_residual_map(fit, include=None, plotter=None):
    """Plot the residual-map of a lens fit.

    Set *autolens.datas.array.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which include a list of every model image, normalized_residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the normalized_residual_map are plotted.
    """
    plotter.plot_array(
        array=fit.normalized_residual_map,
        mask=include.mask_from_fit(fit=fit),
        light_profile_centres=include.light_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        mass_profile_centres=include.mass_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        critical_curves=include.critical_curves_from_obj(obj=fit.tracer),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def chi_squared_map(fit, include=None, plotter=None):
    """Plot the chi-squared-map of a lens fit.

    Set *autolens.datas.array.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which include a list of every model image, residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the chi-squareds are plotted.
    """
    plotter.plot_array(
        array=fit.chi_squared_map,
        mask=include.mask_from_fit(fit=fit),
        light_profile_centres=include.light_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        mass_profile_centres=include.mass_profile_centres_from_obj(
            obj=fit.tracer.image_plane
        ),
        critical_curves=include.critical_curves_from_obj(obj=fit.tracer),
        include_origin=include.origin,
    )



import autofit as af
import autolens as al
from test_autolens.integration.tests.imaging import runner

test_type = "lens__source_inversion"
test_name = "lens_mass__source_rectangular__hyper"
data_name = "lens_sie__source_smooth"
instrument = "vro"


def make_pipeline(name, folders, search=af.DynestyStatic()):

    mass = af.PriorModel(al.mp.EllipticalIsothermal)

    mass.centre.centre_0 = 0.0
    mass.centre.centre_1 = 0.0
    mass.einstein_radius = 1.6

    phase1 = al.PhaseImaging(
        phase_name="phase_1",
        folders=folders,
        galaxies=dict(
            lens=al.GalaxyModel(redshift=0.5, mass=mass),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=al.pix.Rectangular,
                regularization=al.reg.Constant,
            ),
        ),
        search=search,
    )

    phase1.search.const_efficiency_mode = True
    phase1.search.n_live_points = 60
    phase1.search.facc = 0.8

    phase1.extend_with_multiple_hyper_phases(hyper_galaxies_search=True)

    phase2 = al.PhaseImaging(
        phase_name="phase_2",
        folders=folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                mass=phase1.result.model.galaxies.lens.mass,
                hyper_galaxy=al.HyperGalaxy,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=phase1.result.model.galaxies.source.pixelization,
                regularization=phase1.result.model.galaxies.source.regularization,
                hyper_galaxy=phase1.result.hyper_combined.instance.galaxies.source.hyper_galaxy,
            ),
        ),
        search=search,
    )

    phase2.search.const_efficiency_mode = True
    phase2.search.n_live_points = 40
    phase2.search.facc = 0.8

    return al.PipelineDataset(name, phase1, phase2)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import autofit as af
from astropy import cosmology as cosmo
from autogalaxy.galaxy import fit_galaxy
from autogalaxy.galaxy import masked_galaxy_data
from autogalaxy.pipeline.phase import abstract
from autolens.pipeline import visualizer


class Analysis(af.Analysis):
    def __init__(self, cosmology, results, image_path):
        self.cosmology = cosmology
        self.results = results
        self.visualizer = visualizer.PhaseGalaxyVisualizer(image_path)


# noinspection PyAbstractClass
class AnalysisSingle(Analysis):
    def __init__(self, galaxy_data, cosmology, image_path: str, results=None):
        super().__init__(cosmology=cosmology, image_path=image_path, results=results)

        self.galaxy_data = galaxy_data

    def log_likelihood_function(self, instance):
        fit = self.fit_for_instance(instance=instance)
        return fit.figure_of_merit

    def visualize(self, instance, during_analysis):
        fit = self.fit_for_instance(instance=instance)

        self.visualizer.plot_galaxy_fit_subplot(fit)

        if during_analysis:
            self.visualizer.plot_fit_individuals(fit)
        else:

            if self.visualizer.plot_ray_tracing_all_at_end_png:
                self.visualizer.plot_fit_individuals(
                    fit=fit, plot_all=True, image_format="png"
                )

            if self.visualizer.plot_ray_tracing_all_at_end_fits:
                self.visualizer.plot_fit_individuals(
                    fit=fit, plot_all=True, image_format="fits", path_suffix="/fits/"
                )

        return fit

    def fit_for_instance(self, instance):
        """
        Determine the fit of a lens galaxy and source galaxy to the masked_imaging in
        this lens.

        Parameters
        ----------
        instance
            A model instance with attributes

        Returns
        -------
        fit: Fit
            A fractional value indicating how well this model fit and the model
            masked_imaging itself
        """
        return fit_galaxy.FitGalaxy(
            masked_galaxy_dataset=self.galaxy_data, model_galaxies=instance.galaxies
        )


# noinspection PyAbstractClass
class AnalysisDeflections(Analysis):
    def __init__(
        self, galaxy_data_y, galaxy_data_x, cosmology, image_path, results=None
    ):
        super().__init__(cosmology=cosmology, image_path=image_path, results=results)

        self.galaxy_data_y = galaxy_data_y
        self.galaxy_data_x = galaxy_data_x

    def log_likelihood_function(self, instance):
        fit_y, fit_x = self.fit_for_instance(instance=instance)
        return fit_y.figure_of_merit + fit_x.figure_of_merit

    def visualize(self, instance, during_analysis):

        fit_y, fit_x = self.fit_for_instance(instance=instance)

        if self.visualizer.plot_subplot_galaxy_fit:
            self.visualizer.plot_galaxy_fit_subplot(fit_y, path_suffix="/fit_y_")
            self.visualizer.plot_galaxy_fit_subplot(fit_x, path_suffix="/fit_x_")

        if during_analysis:
            self.visualizer.plot_fit_individuals(fit_y, path_suffix="/fit_y")
            self.visualizer.plot_fit_individuals(fit_x, path_suffix="/fit_x")
        else:
            if self.visualizer.plot_ray_tracing_all_at_end_png:
                self.visualizer.plot_fit_individuals(
                    fit_y, path_suffix="/fits/fit_y", plot_all=True
                )
                self.visualizer.plot_fit_individuals(
                    fit_x, path_suffix="/fits/fit_x", plot_all=True
                )

            if self.visualizer.plot_ray_tracing_all_at_end_fits:
                self.visualizer.plot_fit_individuals(
                    fit_y, path_suffix="/fits/fit_y", plot_all=True, image_format="fits"
                )
                self.visualizer.plot_fit_individuals(
                    fit_x, path_suffix="/fits/fit_x", plot_all=True, image_format="fits"
                )

        return fit_y, fit_x

    def fit_for_instance(self, instance):

        fit_y = fit_galaxy.FitGalaxy(
            masked_galaxy_dataset=self.galaxy_data_y, model_galaxies=instance.galaxies
        )
        fit_x = fit_galaxy.FitGalaxy(
            masked_galaxy_dataset=self.galaxy_data_x, model_galaxies=instance.galaxies
        )

        return fit_y, fit_x


class PhaseGalaxy(abstract.AbstractPhase):
    galaxies = af.PhaseProperty("galaxies")

    Analysis = Analysis

    def __init__(
        self,
        phase_name,
        search,
        folders=tuple(),
        galaxies=None,
        use_image=False,
        use_convergence=False,
        use_potential=False,
        use_deflections=False,
        sub_size=2,
        pixel_scales_interp=None,
        cosmology=cosmo.Planck15,
    ):
        """
        A phase in an lens pipeline. Uses the set non_linear search to try to fit
        models and hyper_galaxies passed to it.

        Parameters
        ----------
        search: class
            The class of a non_linear search
        sub_size: int
            The side length of the subgrid
        """

        super(PhaseGalaxy, self).__init__(
            phase_name=phase_name, folders=folders, search=search
        )
        self.cosmology = cosmology
        self.use_image = use_image
        self.use_convergence = use_convergence
        self.use_potential = use_potential
        self.use_deflections = use_deflections
        self.galaxies = galaxies
        self.sub_size = sub_size
        self.pixel_scales_interp = pixel_scales_interp

    def run(self, galaxy_data, mask, info=None, results=None):
        """
        Run this phase.

        Parameters
        ----------
        galaxy_data
        mask: Mask
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            An object describing the results of the last phase or None if no phase has
            been executed

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper_galaxies.
        """
        analysis = self.make_analysis(
            galaxy_data=galaxy_data, results=results, mask=mask
        )

        self.save_metadata(galaxy_data.name)
        self.model = self.model.populate(results)

        result = self.run_analysis(
            analysis=analysis, info=info, pickle_files=pickle_files
        )

        return self.make_result(result, analysis)

    def make_analysis(self, galaxy_data, mask, results=None):
        """
        Create an lens object. Also calls the prior passing and masked_imaging modifying
        functions to allow child classes to change the behaviour of the phase.

        Parameters
        ----------
        galaxy_data
        mask: Mask
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens: Analysis
            An lens object that the non-linear search calls to determine the fit of a
             set of values
        """

        if self.use_image or self.use_convergence or self.use_potential:

            galaxy_data = masked_galaxy_data.MaskedGalaxyDataset(
                galaxy_data=galaxy_data[0],
                mask=mask,
                pixel_scales_interp=self.pixel_scales_interp,
                use_image=self.use_image,
                use_convergence=self.use_convergence,
                use_potential=self.use_potential,
                use_deflections_y=self.use_deflections,
                use_deflections_x=self.use_deflections,
            )

            return AnalysisSingle(
                galaxy_data=galaxy_data,
                cosmology=self.cosmology,
                image_path=self.search.paths.image_path,
                results=results,
            )

        elif self.use_deflections:

            galaxy_data_y = masked_galaxy_data.MaskedGalaxyDataset(
                galaxy_data=galaxy_data[0],
                mask=mask,
                pixel_scales_interp=self.pixel_scales_interp,
                use_image=self.use_image,
                use_convergence=self.use_convergence,
                use_potential=self.use_potential,
                use_deflections_y=self.use_deflections,
                use_deflections_x=False,
            )

            galaxy_data_x = masked_galaxy_data.MaskedGalaxyDataset(
                galaxy_data=galaxy_data[1],
                mask=mask,
                pixel_scales_interp=self.pixel_scales_interp,
                use_image=self.use_image,
                use_convergence=self.use_convergence,
                use_potential=self.use_potential,
                use_deflections_y=False,
                use_deflections_x=self.use_deflections,
            )

            return AnalysisDeflections(
                galaxy_data_y=galaxy_data_y,
                galaxy_data_x=galaxy_data_x,
                cosmology=self.cosmology,
                image_path=self.search.paths.image_path,
                results=results,
            )

    # noinspection PyAbstractClass



import time

import autolens as al
import autolens.plot as aplt
import numpy as np

repeats = 1

workspace_path = "/home/jammy/PycharmProjects/PyAuto/autolens_workspace"
dataset_type = "interferometer"
dataset_name = "lens_sie__source_sersic"
dataset_path = f"{workspace_path}/dataset/{dataset_type}/{dataset_name}"

interferometer = al.Interferometer.from_fits(
    visibilities_path=f"{dataset_path}/visibilities.fits",
    noise_map_path=f"{dataset_path}/noise_map.fits",
    uv_wavelengths_path=f"{dataset_path}/uv_wavelengths.fits",
)

# lens_galaxy = al.Galaxy(
#     redshift=0.5,
#     mass=al.mp.EllipticalIsothermal(
#         centre=(0.4, 0.8), einstein_radius=0.1, elliptical_comps=(0.17647, 0.0)
#     ),
# )

lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mp.EllipticalIsothermal(
        centre=(0.0, 0.0),
        einstein_radius=1.6,
        elliptical_comps=al.convert.elliptical_comps_from(axis_ratio=0.9, phi=45.0),
    ),
    shear=al.mp.ExternalShear(elliptical_comps=(0.0, 0.05)),
)

pixelization = al.pix.VoronoiMagnification(shape=(30, 30))

source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pixelization,
    regularization=al.reg.Constant(coefficient=10.0),
)

mask = al.Mask.circular(shape_2d=(256, 256), pixel_scales=0.05, sub_size=1, radius=3.0)

masked_interferometer = al.MaskedInterferometer(
    interferometer=interferometer,
    real_space_mask=mask,
    visibilities_mask=np.full(
        fill_value=False, shape=interferometer.visibilities.shape
    ),
    transformer_class=al.TransformerNUFFT,
)

print("Number of points = " + str(masked_interferometer.grid.sub_shape_1d) + "\n")
print(
    "Number of visibilities = "
    + str(masked_interferometer.visibilities.shape_1d)
    + "\n"
)

start_overall = time.time()

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

start = time.time()

for i in range(repeats):
    fit = al.FitInterferometer(
        masked_interferometer=masked_interferometer,
        tracer=tracer,
        settings_inversion=al.SettingsInversion(use_linear_operators=True),
    )

print(fit.log_evidence)

diff = time.time() - start
print("Time to compute fit = {}".format(diff / repeats))

aplt.FitInterferometer.subplot_fit_real_space(fit=fit)
aplt.Inversion.reconstruction(inversion=fit.inversion)

start = time.time()

for i in range(repeats):
    inversion = tracer.inversion_interferometer_from_grid_and_data(
        grid=masked_interferometer.grid,
        visibilities=masked_interferometer.visibilities,
        noise_map=masked_interferometer.noise_map,
        transformer=masked_interferometer.transformer,
        settings_inversion=al.SettingsInversion(use_linear_operators=True),
    )

diff = time.time() - start
print("Time to compute inversion = {}".format(diff / repeats))

aplt.FitInterferometer.subplot_fit_real_space(fit=fit)
aplt.Inversion.reconstruction(inversion=fit.inversion)



import os

import autofit as af
import autolens as al
from test_autolens.simulate.imaging import instrument_util

# Get the relative path to the config files and output folder in our autolens_workspace.
test_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output papth
conf.instance = conf.Config(
    config_path=test_path + "config", output_path=test_path + "output"
)

# It is convenient to specify the lens name as a string, so that if the pipeline is applied to multiple images we \
# don't have to change all of the path entries in the function below.
data_name = "no_lens_source_smooth"
instrument = "euclid"

# Setup the size of the sub-grid and mask used for this precision analysis.
sub_size = 2
inner_radius = 0.0
outer_radius = 3.0

# The pixel scale of the interpolation grid, where a smaller pixel scale gives a higher resolution grid and therefore
# more precise interpolation of the sub-grid deflection angles.
pixel_scales_interp = 0.2

imaging = instrument_util.load_test_imaging(
    data_name=data_name, instrument=instrument, psf_shape_2d=(21, 21)
)

# The phase is passed the mask we setup below using the radii specified above.
mask = al.Mask.circular_annular(
    shape=imaging.shape_2d,
    pixel_scales=imaging.pixel_scales,
    inner_radius=inner_radius,
    outer_radius=outer_radius,
)

# Plot the imaging data and mask.
aplt.Imaging.subplot_imaging(imaging=imaging, mask=mask)

# To perform the analysis, we set up a phase using the 'phase' module (imported as 'ph').
# A phase takes our galaxy models and fits their parameters using a non-linear search (in this case, MultiNest).
phase = al.PhaseImaging(
    phase_name="phase_interp",
    folders=[data_name, instrument + "_" + str(pixel_scales_interp)],
    galaxies=dict(lens=al.GalaxyModel(mass=al.EllipticalPowerLaw)),
    galaxies=dict(source=al.GalaxyModel(light=al.lp.EllipticalSersic)),
    search=af.DynestyStatic(),
    pixel_scales_interp=pixel_scales_interp,
)

phase.search.const_efficiency_mode = True
phase.search.n_live_points = 50
phase.search.facc = 0.5

# We run the phase on the image, print the results and plotters the fit.
result = phase.run(dataset=imaging, mask=mask)



import autofit as af
from autoconf import conf
from autogalaxy.pipeline import setup
from autogalaxy.profiles import mass_profiles as mp
from autolens import exc


class SetupPipeline(setup.SetupPipeline):
    def __init__(
        self,
        folders=None,
        hyper_galaxies=False,
        hyper_image_sky=False,
        hyper_background_noise=False,
        hyper_galaxies_search=None,
        inversion_search=None,
        hyper_combined_search=None,
        pixelization=None,
        regularization=None,
        lens_light_centre=None,
        align_light_mass_centre=False,
        align_bulge_disk_centre=False,
        align_bulge_disk_elliptical_comps=False,
        disk_as_sersic=False,
        number_of_gaussians=None,
        no_shear=False,
        lens_mass_centre=None,
        constant_mass_to_light_ratio=False,
        align_light_dark_centre=False,
        align_bulge_dark_centre=False,
        include_smbh=False,
        smbh_centre_fixed=True,
        subhalo_instance=None,
        inversion_pixels_fixed=None,
        evidence_tolerance=None,
    ):
        """The setup of a pipeline, which controls how PyAutoLens template pipelines runs, for example controlling
        assumptions about the lens's light profile bulge-disk model or if shear is included in the mass model.

        Users can write their own pipelines which do not use or require the *SetupPipeline* class.

        This class enables pipeline tagging, whereby the setup of the pipeline is used in the template pipeline
        scripts to tag the output path of the results depending on the setup parameters. This allows one to fit
        different models to a dataset in a structured path format.

        Parameters
        ----------
        pixelization : ag.pix.Pixelization
           If the pipeline uses an *Inversion* to reconstruct the galaxy's light, this determines the
           *Pixelization* used.
        regularization : ag.reg.Regularization
           If the pipeline uses an *Inversion* to reconstruct the galaxy's light, this determines the
           *Regularization* scheme used.
        lens_light_centre : (float, float)
           If input, a fixed (y,x) centre of the lens galaxy is used for the light profile model which is not treated 
           as a free parameter by the non-linear search.
        align_bulge_disk_centre : bool
            If a bulge + disk light model (e.g. EllipticalSersic + EllipticalExponential) is used to fit the galaxy,
            *True* will align the centre of the bulge and disk components and not fit them separately.
        align_bulge_disk_phi : bool
            If a bulge + disk light model (e.g. EllipticalSersic + EllipticalExponential) is used to fit the galaxy,
            *True* will align the rotation angles phi of the bulge and disk components and not fit them separately.
        align_bulge_disk_axis_ratio : bool
            If a bulge + disk light model (e.g. EllipticalSersic + EllipticalExponential) is used to fit the galaxy,
            *True* will align the axis-ratios of the bulge and disk components and not fit them separately.
        disk_as_sersic : bool
            If a bulge + disk light model (e.g. EllipticalSersic + EllipticalExponential) is used to fit the galaxy,
            *True* will use an EllipticalSersic for the disk instead of an EllipticalExponential.
        number_of_gaussians : int
            If a multi-Gaussian light model is used to fit the galaxy, this determines the number of Gaussians.
        no_shear : bool
            If True, shear is omitted from the mass model, if False it is included.
        lens_light_centre : (float, float)
           If input, a fixed (y,x) centre of the lens galaxy is used for the mass profile model which is not treated 
           as a free parameter by the non-linear search.
        align_light_mass_centre : bool
            If True, and the mass model is a decomposed single light and dark matter model (e.g. EllipticalSersic +
            SphericalNFW), the centre of the light and dark matter profiles are aligned.
        align_light_mass_centre : bool
            If True, and the mass model is a decomposed bulge, diskand dark matter model (e.g. EllipticalSersic +
            EllipticalExponential + SphericalNFW), the centre of the bulge and dark matter profiles are aligned.
        hyper_galaxies : bool
            If a hyper-pipeline is being used, this determines if hyper-galaxy functionality is used to scale the
            noise-map of the dataset throughout the fitting.
        hyper_image_sky : bool
            If a hyper-pipeline is being used, this determines if hyper-galaxy functionality is used include the
            image's background sky component in the model.
        hyper_background_noise : bool
            If a hyper-pipeline is being used, this determines if hyper-galaxy functionality is used include the
            noise-map's background component in the model.
        """
        super().__init__(
            folders=folders,
            hyper_galaxies=hyper_galaxies,
            hyper_image_sky=hyper_image_sky,
            hyper_background_noise=hyper_background_noise,
            hyper_galaxies_search=hyper_galaxies_search,
            inversion_search=inversion_search,
            hyper_combined_search=hyper_combined_search,
            pixelization=pixelization,
            regularization=regularization,
            align_bulge_disk_centre=align_bulge_disk_centre,
            align_bulge_disk_elliptical_comps=align_bulge_disk_elliptical_comps,
            disk_as_sersic=disk_as_sersic,
            number_of_gaussians=number_of_gaussians,
            inversion_pixels_fixed=inversion_pixels_fixed,
            evidence_tolerance=evidence_tolerance,
        )

        self.no_shear = no_shear
        self.lens_light_centre = lens_light_centre
        self.lens_mass_centre = lens_mass_centre
        self.align_light_mass_centre = align_light_mass_centre

        if align_light_dark_centre and align_bulge_dark_centre:
            raise exc.SettingsException(
                "In PipelineMassSettings align_light_dark_centre and align_bulge_disk_centre"
                "can not both be True (one is not relevent to the light profile you are fitting"
            )

        self.constant_mass_to_light_ratio = constant_mass_to_light_ratio
        self.align_light_dark_centre = align_light_dark_centre
        self.align_bulge_dark_centre = align_bulge_dark_centre
        self.include_smbh = include_smbh
        self.smbh_centre_fixed = smbh_centre_fixed

        self.subhalo_instance = subhalo_instance

    @property
    def tag(self):
        """Generate the pipeline's overall tag, which customizes the 'setup' folder the results are output to.
        """
        return (
            conf.instance.tag.get("pipeline", "pipeline")
            + self.hyper_tag
            + self.inversion_tag
            + self.align_light_mass_centre_tag
            + self.lens_light_centre_tag
            + self.align_bulge_disk_tag
            + self.disk_as_sersic_tag
            + self.number_of_gaussians_tag
            + self.no_shear_tag
            + self.lens_mass_centre_tag
            + self.align_light_dark_centre_tag
            + self.align_bulge_dark_centre_tag
            + self.include_smbh_tag
            + self.subhalo_centre_tag
            + self.subhalo_mass_at_200_tag
        )

    @property
    def no_shear_tag(self):
        """Generate a tag if an external shear is included in the mass model of the pipeline  are 
        fixedto a previous estimate, or varied during the analysis, to customize pipeline output paths..

        This changes the setup folder as follows:

        no_shear = False -> setup__with_shear
        no_shear = True -> setup___no_shear
        """
        if not self.no_shear:
            return "__" + conf.instance.tag.get("pipeline", "with_shear")
        return "__" + conf.instance.tag.get("pipeline", "no_shear")

    @property
    def lens_light_centre_tag(self):
        """Generate a tag if the lens light model centre of the pipeline is fixed to an input value, to customize 
        pipeline output paths.

        This changes the setup folder as follows:

        lens_light_centre = None -> setup
        lens_light_centre = (1.0, 1.0) -> setup___lens_light_centre_(1.0, 1.0)
        lens_light_centre = (3.0, -2.0) -> setup___lens_light_centre_(3.0, -2.0)
        """
        if self.lens_light_centre is None:
            return ""

        y = "{0:.2f}".format(self.lens_light_centre[0])
        x = "{0:.2f}".format(self.lens_light_centre[1])
        return (
            "__"
            + conf.instance.tag.get("pipeline", "lens_light_centre")
            + "_("
            + y
            + ","
            + x
            + ")"
        )

    @property
    def lens_mass_centre_tag(self):
        """Generate a tag if the lens mass model centre of the pipeline is fixed to an input value, to customize 
        pipeline output paths.

        This changes the setup folder as follows:

        lens_mass_centre = None -> setup
        lens_mass_centre = (1.0, 1.0) -> setup___lens_mass_centre_(1.0, 1.0)
        lens_mass_centre = (3.0, -2.0) -> setup___lens_mass_centre_(3.0, -2.0)
        """
        if self.lens_mass_centre is None:
            return ""

        y = "{0:.2f}".format(self.lens_mass_centre[0])
        x = "{0:.2f}".format(self.lens_mass_centre[1])
        return (
            "__"
            + conf.instance.tag.get("pipeline", "lens_mass_centre")
            + "_("
            + y
            + ","
            + x
            + ")"
        )

    @property
    def constant_mass_to_light_ratio_tag(self):
        """Generate a tag for whether the mass-to-light ratio in a light-dark mass model is constaant (shared amongst
         all light and mass profiles) or free (all mass-to-light ratios are free parameters).

        This changes the setup folder as follows:

        constant_mass_to_light_ratio = False -> mlr_free
        constant_mass_to_light_ratio = True -> mlr_constant
        """
        if self.constant_mass_to_light_ratio:
            return (
                f"__{conf.instance.tag.get('pipeline', 'constant_mass_to_light_ratio')}"
            )
        return f"__{conf.instance.tag.get('pipeline', 'free_mass_to_light_ratio')}"

    @property
    def align_light_mass_centre_tag(self):
        """Generate a tag if the lens mass model is centre is aligned with that of its light profile.

        This changes the setup folder as follows:

        align_light_mass_centre = False -> setup
        align_light_mass_centre = True -> setup___align_light_mass_centre
        """
        if self.lens_light_centre is not None and self.lens_mass_centre is not None:
            return ""

        if not self.align_light_mass_centre:
            return ""
        return "__" + conf.instance.tag.get("pipeline", "align_light_mass_centre")

    @property
    def align_light_dark_centre_tag(self):
        """Generate a tag if the lens mass model is a decomposed light + dark matter model if their centres are aligned.

        This changes the setup folder as follows:

        align_light_dark_centre = False -> setup
        align_light_dark_centre = True -> setup___align_light_dark_centre
        """
        if not self.align_light_dark_centre:
            return ""
        return f"__{conf.instance.tag.get('pipeline', 'align_light_dark_centre')}"

    @property
    def align_bulge_dark_centre_tag(self):
        """Generate a tag if the lens mass model is a decomposed bulge + disk + dark matter model if the bulge centre
        is aligned with the dark matter centre.

        This changes the setup folder as follows:

        align_bulge_dark_centre = False -> setup
        align_bulge_dark_centre = True -> setup___align_bulge_dark_centre
        """
        if not self.align_bulge_dark_centre:
            return ""
        return "__" + conf.instance.tag.get("pipeline", "align_bulge_dark_centre")

    @property
    def include_smbh_tag(self):
        """Generate a tag if the lens mass model includes a _PointMass_ representing a super-massive black hole (smbh).
        
        The tag includes whether the _PointMass_ centre is fixed or fitted for as a free parameter.

        This changes the setup folder as follows:

        include_smbh = False -> setup
        include_smbh = True, smbh_centre_fixed=True -> setup___smbh_centre_fixed
        include_smbh = True, smbh_centre_fixed=False -> setup___smbh_centre_free
        """
        if not self.include_smbh:
            return ""

        include_smbh_tag = conf.instance.tag.get("pipeline", "include_smbh")

        if self.smbh_centre_fixed:

            smbh_centre_tag = conf.instance.tag.get("pipeline", "smbh_centre_fixed")

        else:

            smbh_centre_tag = conf.instance.tag.get("pipeline", "smbh_centre_free")

        return f"__{include_smbh_tag}_{smbh_centre_tag}"

    @property
    def subhalo_centre_tag(self):
        """Generate a tag if the subhalo mass model centre of the pipeline is fixed to an input value, to customize 
        pipeline output paths.

        This changes the setup folder as follows:

        subhalo_centre = None -> setup
        subhalo_centre = (1.0, 1.0) -> setup___sub_centre_(1.0, 1.0)
        subhalo_centre = (3.0, -2.0) -> setup___sub_centre_(3.0, -2.0)
        """
        if self.subhalo_instance is None:
            return ""
        else:
            y = "{0:.2f}".format(self.subhalo_instance.centre[0])
            x = "{0:.2f}".format(self.subhalo_instance.centre[1])
            return (
                "__"
                + conf.instance.tag.get("pipeline", "subhalo_centre")
                + "_("
                + y
                + ","
                + x
                + ")"
            )

    @property
    def subhalo_mass_at_200_tag(self):
        """Generate a tag if the subhalo mass model mass_at_200 of the pipeline is fixed to an input value, to
        customize pipeline output paths.

        This changes the setup folder as follows:

        subhalo_mass_at_200 = None -> setup
        subhalo_mass_at_200 = 1e8 -> setup___sub_mass_1.0e+08
        subhalo_mass_at_200 = 1e9 -> setup___sub_mass_1.0e+09
        """
        if self.subhalo_instance is None:
            return ""
        else:

            return (
                "__"
                + conf.instance.tag.get("pipeline", "subhalo_mass_at_200")
                + "_"
                + "{0:.1e}".format(self.subhalo_instance.mass_at_200_input)
            )

    def smbh_from_centre(self, centre, centre_sigma=0.1):
        """
        Create a _PriorModel_ of a _PointMass_ _MassProfile_ if *include_smbh* is True, which is fitted for in the
        mass-model too represent a super-massive black-hole (smbh).

        The centre of the smbh is an input parameter of the functiono, and this centre is either fixed to the input
        values as an instance or fitted for as a model.

        Parameters
        ----------
        centre : (float, float)
            The centre of the _PointMass_ that repreents the super-massive black hole.
        centre_fixed : bool
            If True, the centre is fixed to the input values, else it is fitted for as free parameters.
        centre_sigma : float
            If the centre is free, this is the sigma value of each centre's _GaussianPrior_.
        """
        if not self.include_smbh:
            return None

        smbh = af.PriorModel(mp.PointMass)

        if self.smbh_centre_fixed:
            smbh.centre = centre
        else:
            smbh.centre.centre_0 = af.GaussianPrior(mean=centre[0], sigma=centre_sigma)
            smbh.centre.centre_1 = af.GaussianPrior(mean=centre[1], sigma=centre_sigma)

        return smbh

    def set_mass_to_light_ratios_of_light_and_mass_profiles(
        self, light_and_mass_profiles
    ):
        """
        For an input list of _LightMassProfile_'s which will represent a galaxy with a light-dark mass model, set all
        the mass-to-light ratios of every light and mass profile to the same value if a constant mass-to-light ratio
        is being used, else keep them as free parameters.

        Parameters
        ----------
        light_and_mass_profiles : [LightMassProfile]
            The light and mass profiles which have their mass-to-light ratios changed.
        """

        if self.constant_mass_to_light_ratio:

            for profile in light_and_mass_profiles[1:]:

                profile.mass_to_light_ratio = light_and_mass_profiles[
                    0
                ].mass_to_light_ratio



import autolens as al


def test__tag__mixture_of_values():

    settings = al.SettingsPhaseImaging(
        settings_masked_imaging=al.SettingsMaskedImaging(
            grid_class=al.Grid,
            grid_inversion_class=al.Grid,
            sub_size=2,
            signal_to_noise_limit=2,
            bin_up_factor=None,
            psf_shape_2d=None,
        ),
        settings_lens=al.SettingsLens(positions_threshold=2.0),
    )

    assert settings.phase_tag_no_inversion == "settings__grid_sub_2__snr_2__pos_2.00"
    assert (
        settings.phase_tag_with_inversion
        == "settings__grid_sub_2_inv_sub_2__snr_2__pos_2.00"
    )

    settings = al.SettingsPhaseImaging(
        settings_masked_imaging=al.SettingsMaskedImaging(
            grid_class=al.GridIterate,
            grid_inversion_class=al.GridInterpolate,
            fractional_accuracy=0.5,
            pixel_scales_interp=0.3,
            bin_up_factor=3,
            psf_shape_2d=(2, 2),
        ),
        settings_lens=al.SettingsLens(auto_positions_factor=0.5),
        settings_pixelization=al.SettingsPixelization(
            use_border=False, is_stochastic=True
        ),
        log_likelihood_cap=200.01,
    )

    assert (
        settings.phase_tag_no_inversion
        == "settings__grid_facc_0.5__bin_3__psf_2x2__auto_pos_x0.50__lh_cap_200.0"
    )
    assert (
        settings.phase_tag_with_inversion
        == "settings__grid_facc_0.5_inv_interp_0.300__bin_3__psf_2x2__auto_pos_x0.50__no_border__stochastic__lh_cap_200.0"
    )

    settings = al.SettingsPhaseInterferometer(
        masked_interferometer=al.SettingsMaskedInterferometer(
            grid_class=al.GridIterate,
            grid_inversion_class=al.GridInterpolate,
            fractional_accuracy=0.5,
            pixel_scales_interp=0.3,
            transformer_class=al.TransformerDFT,
        ),
        settings_pixelization=al.SettingsPixelization(
            use_border=False, is_stochastic=True
        ),
        log_likelihood_cap=100.01,
    )

    assert (
        settings.phase_tag_no_inversion == "settings__grid_facc_0.5__dft__lh_cap_100.0"
    )
    assert (
        settings.phase_tag_with_inversion
        == "settings__grid_facc_0.5_inv_interp_0.300__dft__no_border__stochastic__lh_cap_100.0"
    )

    settings = al.SettingsPhaseInterferometer(
        masked_interferometer=al.SettingsMaskedInterferometer(
            transformer_class=al.TransformerNUFFT
        ),
        settings_inversion=al.SettingsInversion(use_linear_operators=True),
    )

    assert settings.phase_tag_no_inversion == "settings__grid_sub_2__nufft"
    assert (
        settings.phase_tag_with_inversion
        == "settings__grid_sub_2_inv_sub_2__nufft__lop"
    )



from autoconf import conf
from autolens import exc
from autolens.fit import fit_positions

import copy


class SettingsLens:
    def __init__(
        self,
        auto_positions_factor=None,
        auto_positions_minimum_threshold=None,
        positions_threshold=None,
    ):

        self.auto_positions_factor = auto_positions_factor
        self.auto_positions_minimum_threshold = auto_positions_minimum_threshold
        self.positions_threshold = positions_threshold

    @property
    def tag(self):
        return self.auto_positions_factor_tag + self.positions_threshold_tag

    @property
    def auto_positions_factor_tag(self):
        """Generate an auto positions factor tag, to customize phase names based on the factor automated positions are
        required to trace within one another.

        This changes the phase name 'phase_name' as follows:

        auto_positions_factor = None -> phase_name
        auto_positions_factor = 2.0 -> phase_name__auto_pos_x2.00
        auto_positions_factor = 3.0 -> phase_name__auto_pos_x3.00
        """
        if self.auto_positions_factor is None:
            return ""

        if self.auto_positions_minimum_threshold is not None:
            auto_positions_minimum_threshold_tag = f"_{conf.instance.tag.get('lens', 'auto_positions_minimum_threshold')}_{str(self.auto_positions_minimum_threshold)}"
        else:
            auto_positions_minimum_threshold_tag = ""

        return (
            "__"
            + conf.instance.tag.get("lens", "auto_positions_factor")
            + "_x{0:.2f}".format(self.auto_positions_factor)
            + auto_positions_minimum_threshold_tag
        )

    @property
    def positions_threshold_tag(self):
        """Generate a positions threshold tag, to customize phase names based on the threshold that positions are required \
        to trace within one another.

        This changes the phase name 'phase_name' as follows:

        positions_threshold = 1 -> phase_name
        positions_threshold = 2 -> phase_name_positions_threshold_2
        positions_threshold = 2 -> phase_name_positions_threshold_2
        """

        old_tag = conf.instance.general.get("tag", "old_tag", bool)

        if old_tag and self.auto_positions_factor is not None:
            return ""

        if self.positions_threshold is None:
            return ""
        return (
            "__"
            + conf.instance.tag.get("lens", "positions_threshold", str)
            + "_{0:.2f}".format(self.positions_threshold)
        )

    def check_positions_trace_within_threshold_via_tracer(self, positions, tracer):

        if not tracer.has_mass_profile or len(tracer.planes) == 1:
            return

        if positions is not None and self.positions_threshold is not None:

            positions_fit = fit_positions.FitPositionsSourcePlaneMaxSeparation(
                positions=positions, tracer=tracer, noise_value=1.0
            )

            if not positions_fit.maximum_separation_within_threshold(
                self.positions_threshold
            ):
                raise exc.RayTracingException

    def modify_positions_threshold(self, positions_threshold):

        settings = copy.copy(self)
        settings.positions_threshold = positions_threshold
        return settings



import autofit as af
import autolens as al

"""
All pipelines begin with a comment describing the pipeline and a phase-by-phase description of what it does.

In this pipeline, we fit the a strong lens using an _EllipticalSersic_ _LightProfile_, _EllipticalIsothermal_ 
_MassProfile_ and parametric _EllipticalSersic_ source.

The pipeline is three phases:

Phase 1:

    Fit and subtract the lens light model.
    
    Lens Light: EllipticalSersic
    Lens Mass: None
    Source Light: None
    Prior Passing: None
    Notes: None

Phase 2:

    Fit the lens mass model and source _LightProfile_.
    
    Lens Light: EllipticalSersic
    Lens Mass: EllipticalIsothermal + ExternalShear
    Source Light: EllipticalSersic
    Prior Passing: Lens Light (instance -> phase 1).
    Notes: Uses the lens subtracted image from phase 1.

Phase 3:

    Refine the lens light and mass models and source light model using priors initialized from phases 1 and 2.
    
    Lens Light: EllipticalSersic
    Lens Mass: EllipticalIsothermal + ExternalShear
    Source Light: EllipticalSersic
    Prior Passing: Lens light (model -> phase 1), lens mass and source light (model -> phase 2).
    Notes: None
"""


def make_pipeline(setup, settings):

    pipeline_name = "pipeline__light_and_source"

    """
    A pipelines takes the 'folders' as input, which together with the pipeline name specify the path structure 
    of the output. In the pipeline runner we pass the folders ["howtolens", c3_t1_lens_and_source], making the
    output of this pipeline 'autolens_workspace/output/howtolens/c3_t1_lens_and_source/pipeline__light_and_source'.

    The output path is also tagged according to the _SetupPipeline_, in an analagous fashion to how the 
    _SettingsPhaseImaging_ tagged the output paths of phases. In this example, we do not use an _ExternalShear_
    in the mass model, and the pipeline is tagged accordingly.
    """

    setup.folders.append(pipeline_name)
    setup.folders.append(setup.tag)

    """
    Phase 1: Fit only the lens galaxy's light, where we:

        1) Set priors on the lens galaxy (y,x) centre such that we assume the image is centred around the lens galaxy.

    We create the phase using the same notation as in chapter 2. Note how we are using the 'fast' _Dynesty_ settings
    covered in chapter 2.
    """

    phase1 = al.PhaseImaging(
        phase_name="phase_1__lens_sersic",
        folders=setup.folders,
        galaxies=dict(lens=al.GalaxyModel(redshift=0.5, light=al.lp.EllipticalSersic)),
        settings=settings,
        search=af.DynestyStatic(n_live_points=30, evidence_tolerance=5.0),
    )

    """
    Phase 2: Fit the lens galaxy's mass and source galaxy's light, where we:

        1) Fix the foreground lens light subtraction to the lens galaxy light model from phase 1.
        2) Set priors on the centre of the lens galaxy's _MassProfile_ by linking them to those inferred for 
           the _LightProfile_ in phase 1.
           
    In phase 2, we fit the source-galaxy's light. Thus, we want to fix the lens light model to the model inferred
    in phase 1, ensuring the image we fit is lens subtracted. We do this below by passing the lens light as an
 'instance' object, a trick we use in nearly all pipelines!

    By passing an 'instance', we are telling __PyAutoLens__ that we want it to pass the maximum log likelihood result of
    that phase and use those parameters as fixed values in the model. The model parameters passed as an 'instance' are
    not free parameters fitted for by the non-linear search, thus this reduces the dimensionality of the non-linear 
    search making model-fitting faster and more reliable. 
     
    Thus, phase2 includes the lens light model from phase 1, but it is completely fixed during the model-fit!
    """

    mass = af.PriorModel(al.mp.EllipticalIsothermal)
    mass.centre_0 = phase1.result.model.galaxies.lens.light.centre_0
    mass.centre_1 = phase1.result.model.galaxies.lens.light.centre_1

    phase2 = al.PhaseImaging(
        phase_name="phase_2__lens_sie__source_sersic",
        folders=setup.folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=phase1.result.instance.galaxies.lens.light,
                mass=mass,
            ),
            source=al.GalaxyModel(redshift=1.0, light=al.lp.EllipticalSersic),
        ),
        settings=settings,
        search=af.DynestyStatic(n_live_points=50, evidence_tolerance=5.0),
    )

    """
    Phase 3: Fit simultaneously the lens and source galaxies, where we:

        1) Set the lens's light, mass, and source's light using the results of phases 1 and 2.
        
    As in chapter 2, we can use the 'model' attribute to do this. Our _Dynesty_ search now uses slower and more 
    thorough settings than the previous phases, to ensure we robustly quantify the errors.
    """

    phase3 = al.PhaseImaging(
        phase_name="phase_3__lens_sersic_sie__source_sersic",
        folders=setup.folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=phase1.result.model.galaxies.lens.light,
                mass=phase2.result.model.galaxies.lens.mass,
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=phase2.result.model.galaxies.source.light
            ),
        ),
        settings=settings,
        search=af.DynestyStatic(n_live_points=100),
    )

    return al.PipelineDataset(pipeline_name, phase1, phase2, phase3)



import numpy as np
from autolens.plot import plotters

# Although we could test_autoarray the deflection angles without using an image (e.al. by just making a grid), we have chosen to
# set this test_autoarray up using an image and mask. This gives run-time numbers that can be easily related to an actual lens
# analysis

sub_size = 2
inner_radius = 0.2
outer_radius = 4.0

print("sub grid size = " + str(sub_size))
print("annular inner mask radius = " + str(inner_radius) + "\n")
print("annular outer mask radius = " + str(outer_radius) + "\n")

for instrument in ["hst_up"]:

    print()

    imaging = instrument_util.load_test_imaging(
        data_name="lens_sie__source_smooth", instrument=instrument, psf_shape_2d=(3, 3)
    )
    mask = al.Mask.circular_annular(
        shape=imaging.shape_2d,
        pixel_scales=imaging.pixel_scales,
        inner_radius=inner_radius,
        outer_radius=outer_radius,
    )
    masked_imaging = al.MaskedImaging(imaging=imaging, mask=mask, sub_size=sub_size)

    print("Deflection angle run times for image type " + instrument + "\n")
    print("Number of points = " + str(masked_imaging.grid.sub_shape_1d) + "\n")

    interpolator = al.GridInterpolate.from_mask_grid_and_pixel_scales_interps(
        mask=masked_imaging.mask, grid=masked_imaging.grid, pixel_scales_interp=0.05
    )

    print(
        "Number of interpolation points = "
        + str(interpolator.grid_interp.sub_shape_1d)
        + "\n"
    )

    ### EllipticalIsothermal ###

    mass_profile = al.EllipticalIsothermal(
        centre=(0.0, 0.0), elliptical_comps=(0.111111, 0.0), einstein_radius=1.0
    )

    interp_deflections = mass_profile.deflections_from_grid(
        grid=interpolator.grid_interp
    )
    deflections = np.zeros((masked_imaging.grid.sub_shape_1d, 2))
    deflections[:, 0] = interpolator.interpolated_array_from_array_interp(
        array_interp=interp_deflections[:, 0]
    )
    deflections[:, 1] = interpolator.interpolated_array_from_array_interp(
        array_interp=interp_deflections[:, 1]
    )

    true_deflections = mass_profile.deflections_from_grid(grid=masked_imaging.grid)

    true_deflections_y_2d = masked_imaging.grid.array_stored_1d_from_sub_array_1d(
        sub_array_1d=true_deflections[:, 0]
    )
    true_deflections_x_2d = masked_imaging.grid.array_stored_1d_from_sub_array_1d(
        sub_array_1d=true_deflections[:, 1]
    )

    difference_y = deflections[:, 0] - true_deflections[:, 0]
    difference_x = deflections[:, 1] - true_deflections[:, 1]

    print("interpolation y error: ", np.mean(difference_y))
    print("interpolation y uncertainty: ", np.std(difference_y))
    print("interpolation y max error: ", np.max(difference_y))
    print("interpolation x error: ", np.mean(difference_x))
    print("interpolation x uncertainty: ", np.std(difference_x))
    print("interpolation x max error: ", np.max(difference_x))

    difference_y_2d = masked_imaging.grid.array_stored_1d_from_sub_array_1d(
        sub_array_1d=difference_y
    )
    difference_x_2d = masked_imaging.grid.array_stored_1d_from_sub_array_1d(
        sub_array_1d=difference_x
    )

    aplt.Array(array=true_deflections_y_2d)
    aplt.Array(array=difference_y_2d)

    aplt.Array(array=true_deflections_x_2d)
    aplt.Array(array=difference_x_2d)

    # difference_percent_y = (np.abs(difference_y) / np.abs(true_deflections[:,0]))*100.0
    # difference_percent_x = (np.abs(difference_x) / np.abs(true_deflections[:,1]))*100.0
    #
    # print("interpolation y mean percent difference: ", np.mean(difference_percent_y))
    # print("interpolation y std percent difference: ", np.std(difference_percent_y))
    # print("interpolation y max percent difference: ", np.max(difference_percent_y))
    # print("interpolation x mean percent difference: ", np.mean(difference_percent_x))
    # print("interpolation x std percent difference: ", np.std(difference_percent_x))
    # print("interpolation x mean percent difference: ", np.max(difference_percent_x))
    #
    # difference_percent_y_2d = masked_imaging.grid.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
    #     sub_array_1d=difference_percent_y)
    # difference_percent_x_2d = masked_imaging.grid.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
    #     sub_array_1d=difference_percent_x)
    #
    # aplt.Array(arrays=difference_percent_y_2d)
    # aplt.Array(arrays=difference_percent_x_2d)
